// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//




/*

Uses code by Hans Muller and Kathy Walrath from
http://java.sun.com/products/jfc/tsc/articles/threads/threads2.html

*/


package de.uka.ilkd.key.gui;

import java.util.HashSet;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.util.Debug;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;

/**
 * Applies heuristic Taclets in an own thread.
 */
public class ApplyHeuristics {

    private TacletApp lastApp=null;
    private InteractiveProver iaProver;
    /** the proof that is worked with */ 
    private Proof proof;
    /** the maximum of allowed heuristics applications */
    private int maxApplications;
    /** the java information object */
    private JavaInfo javaInfo;
    /** list of goals on which the heuristics should be applied */
    private ListOfGoal goalList;
    /** set of goals where no rule is applicable at the current run */
    private HashSet nonApplicableGoals;
    
    private SwingWorker worker;
    
    private JFrame stopper;
    private JProgressBar progressBar;

    /** number of rules automatically applied */
    private int countApplied = 0;
    private long time;
    
    /** interrupted by the user? */
    private boolean userTerminated = false;


    /**
     * creates a new ApplyHeuristics with a given IaProver that provides
     * the necessary data and notification mechanisms and with a given
     * StopHeuristics, i.e. a (graphical) user dialog that may stop this
     * ApplyHeuristics and may itself be finished by this class.
     * @param iaProver the InteractiveProver with the necessary data and methods
     * @param maxApp an int bounding the number of heuristics
     * applications before an interactive step is required
     */
    public ApplyHeuristics(InteractiveProver iaProver, Proof proof, int maxApp) {
	this.iaProver = iaProver;
	this.proof = proof;
	javaInfo = iaProver.mediator().getJavaInfo();
	maxApplications = maxApp;
        
        stopper = new JFrame("Processing Heuristics");
        progressBar = new JProgressBar(0, maxApp-1);
        
        
        // now lay out the stopper frame
        stopper.getContentPane().setLayout(new BorderLayout());
	stopper.getContentPane().add(
           new JLabel("Press Stop to interrupt the execution of heuristics."), 
           BorderLayout.CENTER);

        progressBar.setValue(0);
	progressBar.setStringPainted(true);
	stopper.getContentPane().add(progressBar);

 	JButton okButton = new JButton("Stop");
	okButton.addActionListener(interruptListener); // see below
	JPanel buttonPanel = new JPanel();
 	buttonPanel.add(okButton);
	stopper.getContentPane().add(buttonPanel, BorderLayout.SOUTH);
	stopper.pack();
	stopper.setVisible(true);
    }


    /** applies Taclets that belongs to an active heuristic 
     * @return true iff a rule has been applied, false otherwise
     */
    public synchronized boolean applyHeuristicTaclet() {
	IteratorOfGoal it = goalList.iterator();
	while (it.hasNext()) {
	    Goal goal = it.next();
	    int index = 0;
	    IteratorOfConstrainedFormula fmlIt = goal.seq().iterator();
	    while (fmlIt.hasNext() && !nonApplicableGoals.contains(goal)) {
		PosInOccurrence pos = new PosInOccurrence(fmlIt.next(),  
							  PosInTerm.TOP_LEVEL,
							  goal.seq());
		IteratorOfTacletApp tacletAppIt = 
		    goal.ruleAppIndex()
		    .getTacletAppAtAndBelow(iaProver.activeHeuristics(),
					  false,
					  pos, javaInfo).iterator(); 
		while (tacletAppIt.hasNext()) {
		    TacletApp app = tacletAppIt.next();
		    if (!actuallyContains(goal.appliedRuleApps(),app)) {
			ListOfSVInstantiations svilist 
			    = (app.taclet()).matchIf(goal.node().sequent(), 
						   app.instantiations(), javaInfo);
			IteratorOfSVInstantiations svit=svilist.iterator();
			while (svit.hasNext()) {
			    app=app.setInstantiation(svit.next());
			    if (!app.complete()) {
				app = app.tryToInstantiate(iaProver.mediator().
                                                           getJavaInfo());
			    }
			    if (app!=null) {
				//				if (lastApp!=null && 
				//				    lastApp.taclet().equals(app.taclet()) &&
				//				    lastApp.instantiations().equals
				//				    (app.instantiations())) {
				// This is because replacing an existing formula by
				// an equivalent yields a different TacletApp.
				// Changing .equals of PosInOcc (== for 
				// ConstrainedFormula) would block the continuing 
				// application of the same formula, but would have
				// other effects...
				    //				} else 
				    if (!actuallyContains
					   (goal.appliedRuleApps(),app)){ 
				    Debug.out("\n ==== "
					      +"AUTOMATIC Application ==== ", 
					      app);
				    
				    ListOfGoal newGoals=
					app.execute(goal, 
						    iaProver.mediator().getJavaInfo());
				    lastApp=null;
				    updateGoalList(goal, newGoals);
				    return true;				    
				}
			    }			
			}
		    }
		}
	    }	    
	    // if here no rules have been applicable in goal
	    nonApplicableGoals.add(goal);
	}
	return false;
    }

    private boolean actuallyContains(ListOfRuleApp list, TacletApp rapp) {
	//	if (rapp.taclet().name().toString().equals("auto_insert_eq")) System.out.println("YYY "+list);
	return list.contains(rapp);
// 	IteratorOfRuleApp it=list.iterator();
// 	while (it.hasNext()) {
// 	    TacletApp cmp=(TacletApp)it.next();
// 	    if ((rapp.getClass()==cmp.getClass()) &&
// 		(rapp.taclet().equals(cmp.taclet()))) {
// 		IteratorOfSchemaVariable svit=cmp.instantiations().svIterator();
// 		boolean ok=true;
// 		while (svit.hasNext() && ok) {		    
// 		    SchemaVariable sv=svit.next();
// 		    Object o1=rapp.instantiations().getInstantiation(sv);
// 		    Object o2=cmp.instantiations().getInstantiation(sv);
// 		    if (o1==null || !o1.equals(o2)) {
// 			ok=false;
// 		    }
// 		}
// 		if (ok) return true; else if (cmp.taclet().name().toString().equals("greater")) System.out.println("XXXXXXX\n"+cmp.instantiations()+"\n vs\n"+rapp.instantiations());		    		
// 	    }
// 	}
// 	return false;								   	
    }


    

    private void updateGoalList(Goal goal, 
				ListOfGoal newGoals) {
	proof.replace(goal, newGoals);
	goalList = goalList.removeAll(goal);
	goalList = goalList.prepend(newGoals);
    }


    
    void updateStatus() {
        Runnable doSetProgressBarValue = new Runnable() {
            public void run() {
                progressBar.setValue(countApplied);
            }
        };
        SwingUtilities.invokeLater(doSetProgressBarValue);
    }
    

    /**
     * applies heuristics until this is no longer
     * possible or the thread is interrupted.
     */
    Object doWork() {
        time = System.currentTimeMillis();
        try{
	   Debug.out("Heuristics started.");
	   nonApplicableGoals = new HashSet();
	   while (countApplied < maxApplications &&
	          applyHeuristicTaclet()) {
               countApplied++;
               updateStatus();
               if (Thread.interrupted()) throw new InterruptedException();
	   }
        }
        catch (InterruptedException e) {
            return "Interrupted";  // SwingWorker.get() returns this
        }
        finally{
           time = System.currentTimeMillis()-time;
	   Debug.out("Heuristics stopped.");
	   Debug.out("Applied "+countApplied);
	   Debug.out("Time elapsed: "+ time);
        }
        return "Done.";
    }
    
    
    
    public void start(ListOfGoal goals) {
	goalList = goals;

        /* Invoking start() on the SwingWorker causes a new Thread
         * to be created that will call construct(), and then
         * finished().  Note that finished() is called even if
         * the worker is interrupted because we catch the
         * InterruptedException in doWork().
         */
        worker = new SwingWorker() {
            public Object construct() {
                return doWork();
            }
            public void finished() {
//                    statusField.setText(get().toString());
	        iaProver.fireAutoModeStopped(new ProofEvent(iaProver));

	        // inform that heuristics have been stopped
	        String message = "Heuristics stopped"+
                                 (userTerminated ? "by user.\n" : ".\n");
	        if (countApplied == maxApplications) {
		    message = "Maximal number of heuristic applications" +
                              "has been reached.\n";
	        }
	        // Show message ...
                int closed = iaProver.mediator().getNrGoalsClosedByHeuristics();
	        if (closed>0) {
		    message += closed+
                               ((closed>1) ? " goals have" : " goal has") +
                               " been closed.\n";
	        }
		message += "Applied " + countApplied + " taclets in " +
			   (time/1000)+"."+((time%1000)/100)+" seconds.";
		if (iaProver.getProof().openGoals().size()==0) {
		    message="Proved";
		}
		if ((closed>0) || (countApplied == maxApplications) 
		    || "Proved".equals(message)) {
                   JOptionPane.showMessageDialog(stopper, message, "Note ...", 
		      JOptionPane.INFORMATION_MESSAGE);
                }
		iaProver.mediator().resetNrGoalsClosedByHeuristics();		
                stopper.dispose();
            }
        };
        iaProver.fireAutoModeStarted(new ProofEvent(iaProver));
        worker.start();
    }



    /**
     * This action listener, called by the "Stop" button, interrupts
     * the worker thread which is running this.doWork().  Note that
     * the doWork() method handles InterruptedExceptions cleanly.
     */
    ActionListener interruptListener = new ActionListener() {
        public void actionPerformed(ActionEvent event) {
            userTerminated = true;
            worker.interrupt();
        }
    };
    
    


}
