// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** the class is used to store abbreviations for epsilon terms and the current
 * state of the term. This means if the abbreviation is active or not 
*/
package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.util.pp.UnbalancedBlocksException;
import java.io.IOException;

class EpsilonPrintInfo extends Notation {
    private String abbrev;
    private int index;
    private boolean isAbbreviated;

    /** creates an epsilon print info object
     * @param not the NotationInfo where all EpsilonPrintInfos are stored
     * @param eps the Term 
     */
    public EpsilonPrintInfo(NotationInfo notInfo, Term eps) {
	super(60);
	if (eps.op()!=Op.EPSILON) {
	    throw new RuntimeException("EpsilonPrintInfo expects an epsilon"+
				       " term"); 
	} 
	index = -1;
	isAbbreviated = true;	
	createAbbreviation(notInfo, eps);
    }

    /** creates an abbreviation for an epsilon term. The text representation of
     * terms in a sequent should be different for different terms, that is why
     * the sequent is needed
     * @param not the NotationInfo where all EpsilonPrintInfos are stored
     * @param eps the Term 
     * @return the abbreviation as String
     */ 
    private void createAbbreviation(NotationInfo notationInfo, Term eps) {
	abbrev = "eps "+
	    eps.varsBoundHere(0).getQuantifiableVariable(0).name(); 
	while (notationInfo.containsAbbrev(this)) {
	    index ++; 
	}
	setName(abbrev+(index < 0 ? "" : ""+index)+"...");
    }

    /** looks if two epsilon print infos have the same textual representation */
    public boolean equals(Object obj) {
	if (!(obj instanceof EpsilonPrintInfo)) {
	    return false;
	}
	EpsilonPrintInfo info = (EpsilonPrintInfo) obj;
	return info.getAbbrev().equals(getAbbrev());
    }
    
    public String toString() {
	return getAbbrev();
    }

    /** returns the abbreviation of the epsilon term 
     * @return  the abbreviation of the epsilon term 
     */
    protected String getAbbrev() {
	return abbrev;
    }

    /** true iff epsilon term is abbreviated 
     * @return true iff epsilon term is abbreviated 
     */
    public boolean isAbbreviated() {
	return isAbbreviated;
    }

    /** sets the boolean value depending if epsilon term should be abbreviated
     * or not
     * @param abbreviate a boolean true if abbreviation is wanted
     */
    public void setAbbreviated(boolean abbreviate) {
	isAbbreviated = abbreviate;
    }

    /** sets the new abbreviation of the epsilon term
     * @param name the new Name of the epsilon term
     */
    public void setName(String name) {
        abbrev = name;
    }

    public void print(Term t,SequentPrinter sp) 
	throws UnbalancedBlocksException,IOException {
	sp.printConstant("["+getAbbrev()+"]");
    }

}
