// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;


/** this class implements a persistent constraint. The constraint
 * contains pairs of Metavariables X and Terms t meaning X=t. It
 * offers services like joining two Constraint objects, adding new
 * constraints to this one by unfying two terms and creating all
 * necessary Metavariable - Term pairs. There are no public
 * constructors to build up a new Constraint use the BOTTOM constraint
 * of the Constraint interface (final static class variable) and add
 * the needed constraints. If a constraint would not be satisfiable
 * (cycles, unification failed) the Constraint TOP of interface
 * Constraint is returned.
 */

public class EqualityConstraint implements Constraint {
    
    /** contains a boolean value */
    private final static BooleanContainer
	CONSTRAINTBOOLEANCONTAINER=new BooleanContainer();
    /** used by method unifyhlp to encapsulate a second return value */ 
    private final static BooleanContainer
	unifyable=new BooleanContainer();
    /** stores constraint content as a mapping from Metavariable to
     * Term */
    private MapFromMetavariableToTerm
	map=MapAsListFromMetavariableToTerm.EMPTY_MAP; 
    /** Don't use this constructor, use Constraint.BOTTOM instead */
    public EqualityConstraint() {}

    /** creates a new constraint with map and the new pair mv,
     * term. It has to be ensured that new map satisfies cyclefree.
     */
    private EqualityConstraint(MapFromMetavariableToTerm map, 
			       Metavariable mv,Term t) {
	this.map=map.put(mv,t);
    } 

    private Object readResolve() 
	throws java.io.ObjectStreamException {
	if (isBottom()) {
	    return Constraint.BOTTOM;
	}	
	return this;
    }
    
    /** returns true if Bottom
     * @return true if Bottom 
     */
    public boolean isBottom() {
	return (map == MapAsListFromMetavariableToTerm.EMPTY_MAP);
    }

    /** a constraint being instance of this class is satisfiable. If a
     * method realizes that an unsatisfiable Constraint would be built
     * because of failed unification, cycle or s.th. similar it
     * returns the singleton TOP being instance of the subclass Top
     * @return true always 
     */
    public boolean isSatisfiable() {
	return true;
    }

    /** if one of the terms t1 or t2 is a Metavariable, then they are
     * added to the pairList, else unification continuous with the
     * subterms 
     * ATTENTION:there is a second return value that indicates if
     * unification was successfull. The value is stored in the
     * BooleanContainer unifyable that is a final static class
     * variable. It is true iff unification was successful.
     * @param t1 one of the terms that has to be unified
     * @param t2 the other term 
     * @return list of pairs of Metavariables and Terms that has to be
     * added to the constraint because of the unification of t1 and t2 
     */
    private ListOfEntryOfMetavariableAndTerm 
	unifyHlp(Term t1, Term t2, 
		 ListOfEntryOfMetavariableAndTerm pairList) { 

	ListOfEntryOfMetavariableAndTerm newPairList=pairList;
	unifyable.setVal(true);

	if (t1.op() instanceof Metavariable) {
	    newPairList=pairList.prepend(new
		EntryMV_T((Metavariable)t1.op(),t2)); 
	} else if (t2.op() instanceof Metavariable) {
	    newPairList=pairList.prepend(new
		EntryMV_T((Metavariable)t2.op(),t1)); 
	} else {
	    if (!(t1.op().equals(t2.op()))) {
		unifyable.setVal(false);
		return null; // means TOP
	    }
	    for (int i=0;i<t1.arity();i++) {
		newPairList=unifyHlp(t1.sub(i),t2.sub(i),newPairList);
		if (!unifyable.val()) {
		    return null; // means TOP
		}
	    }
	}
	return newPairList;
    } 

    /** executes unification for terms t1 and t2
     * PRECONDITION: sorts of the given terms are equal 
     * @param t1 Term to be unified 
     * @param t2 term to be unified
     * @return TOP if not possible, else a new constraint with after
     * unification of t1 and t2
     */
    public Constraint unify(Term t1, Term t2) {
	Debug.assert(t1.sort()==t2.sort(),"Sorts of terms are not equal.");
	return unify(t1, t2, CONSTRAINTBOOLEANCONTAINER);
    }

        
    /** executes unification for terms t1 and t2. 
     * PRECONDITION: sorts of the given terms are equal. 
     * @param t1 Term to be unfied 
     * @param t1 Term to be unfied
     * @param subsumed  true iff the new constraint is subsumed by
     * this one
     * @return TOP if not possible, else a new constraint with after
     * unification of t1 and t2 ( == this iff this subsumes the unification )
     */
    public Constraint unify(Term t1, Term t2, BooleanContainer subsumed) {
	Debug.assert(t1.sort()==t2.sort(),"Sorts of terms are not equal.");
	ListOfEntryOfMetavariableAndTerm
	    pairList=SLListOfEntryOfMetavariableAndTerm.EMPTY_LIST; 
	pairList=unifyHlp(t1,t2,pairList);  
	if (!unifyable.val()) {
	    return TOP;
	}

	IteratorOfEntryOfMetavariableAndTerm it     = pairList.iterator();
	Constraint                           result = this;
	EntryOfMetavariableAndTerm           entry;
	Metavariable                         mv;
	Term                                 t;
	Constraint                           res;
	subsumed.setVal(true);

	while ( it.hasNext () && result != TOP ) {
	    entry = it.next();
	    mv    = entry.key();
	    t     = entry.value();
	    res   = ((EqualityConstraint)result).normalize(mv,t);
	    if (res!=result) {
		subsumed.setVal(false);
	    }
	    result=res;	    
	}	
	return result;
    }

    /** checks for cycles and adds additonal constraints if necessary 
     * @param mv Metavariable asked to be mapped to the term t
     * @param t the Term the metavariable should be mapped to (if there
     * are no cycles )
     * @return the resulting Constraint ( == this iff this subsumes
     * the new constraint )
     */ 
    private Constraint normalize(Metavariable mv, Term t) {
	// remove MV cycles
	if (!hasMVCycle(mv,t)) {
	    if (hasCycle(mv,t)) {// cycle
		return TOP;
	    } else if (isDefined(mv)) {
		return unify(valueOf(mv),t);
	    } else {
		return new EqualityConstraint(map,mv,t);
	    }
	}
	return this;
    }

    /** checks equality of constraints by subsuming relation
     */
    public boolean equals ( Object obj ) {
	if ( !( obj instanceof EqualityConstraint ) ) 
	    return obj.equals ( this );
	else {
	    EqualityConstraint c = (EqualityConstraint)obj;
	    return ( c.join ( this ) == c ) && ( this.join ( c ) == this );
	}
    }

    /** joins the given constraint with this constraint
     * and returns the joint new constraint. 
     * @param co Constraint to be joined with this one
     * @return the joined constraint 
     */	
    public Constraint join(Constraint co) { 
	return join(co, CONSTRAINTBOOLEANCONTAINER); 
    }
    

    /** joins constraint co with this constraint
     * and returns the joint new constraint. The BooleanContainer is
     * used to wrap a second return value and indicates a subsumption
     * of co by this constraint. 
     * @param p_co Constraint to be joined with this one
     * @param p_subsuemd the BooleanContainers value set true, iff this
     * constraint subsumes co 
     * @return the joined constraint     
     */
    public Constraint join ( Constraint p_co, BooleanContainer p_subsumed ) {

	if ( !( p_co instanceof EqualityConstraint ) ) 
	    return p_co.join ( this, p_subsumed );
	else {

	    IteratorOfEntryOfMetavariableAndTerm it     =
		((EqualityConstraint)p_co).map.entryIterator();
	    Constraint                           result = this;
	    EntryOfMetavariableAndTerm           entry;
	    Metavariable                         mv;
	    Constraint                           res;

	    p_subsumed.setVal(true);
	    while ( it.hasNext () && result != TOP ) {
		entry  = it    .next      ();
		mv     = entry .key       ();
		res    = ((EqualityConstraint)result).normalize ( mv, entry.value () );  
		if ( res != result ) {
		    p_subsumed.setVal(false);
		}
		result=res;
	    }
	    return result;

	}

    }
    

    /** checks if there is a cycle if the metavariable mv and Term
     * term would be added to this constraint e.g. X=g(Y), Y=f(X) 
     * @param mv the Metavariable   
     * @param term The Term 
     * @return a boolean that is true iff. adding a mapping (mv,term)
     * would cause a cycle 
     */

    boolean hasCycle(Metavariable mv, Term term) {
	HashMapFromMetavariableToMetavariable body=new HashMapFromMetavariableToMetavariable();
	Term checkForCycle=term;
	ListOfTerm fringe=SLListOfTerm.EMPTY_LIST;
	fringe=fringe.prepend(checkForCycle);

	while (fringe!=SLListOfTerm.EMPTY_LIST) {
	    checkForCycle=fringe.head();
	    fringe=fringe.tail();
	    MVCollector coll=new MVCollector();
	    checkForCycle.execPostOrder(coll);
	    IteratorOfMetavariable it=coll.mv();
	    while (it.hasNext()) {
		Metavariable termMV=(Metavariable)it.next();
		Term t=null;
		if (!body.containsKey(termMV)) {
		    if ((t=map.get(termMV))!=null) {
			fringe=fringe.prepend(t);
			body.put(termMV,termMV);
		    }
		    if (termMV.equals(mv)) {
			return true;
			
		    }
		}
	    }
	}
	return false;	
    }
    
    /** checks if the given Metavariables are part of a cycle of metavariables.
     * @param mv Metavariable that shall be mapped to mv2 
     * @param mv2  the Metavariable mv is mapped to 
     * @return a boolean being true iff mapping (mv,mv2) would create
     * a Metavariablecycle
     */
    boolean hasMVCycle(Metavariable mv, Metavariable mv2) {
	if (mv2.equals(mv)) {
	    return true;
	}
	Term t=map.get(mv2);
	if (t==null) {
	    return false;
	}
	if (!(t.op() instanceof Metavariable)) return false;
	Metavariable mv2new=(Metavariable)t.op();
	return hasMVCycle(mv,mv2new);
    }

    /** checks if a metavariable cycle is created if the mapping
     * (mv,t) would be added to this constraint e.g. X=Y, Y=Z, Z=X 
     * @param mv the Metavariable mapped to Term t
     * @param t the Term mv is mapped to
     * @return true iff adding the maping would create a Metavariablecycle 
     */   
    boolean hasMVCycle(Metavariable mv, Term t) {
	if (t==null || !(t.op() instanceof Metavariable)) return false;
	return hasMVCycle(mv,(Metavariable)t.op())||
	    hasMVCycle((Metavariable)t.op(),map.get((Metavariable)t.op()));
    }
        
    /**ONLY FOR TESTS DONT USE THEM IN ANOTHER WAY
     * @return true if metavar is contained as key */
    boolean isDefined(Metavariable mv) {
	return map.containsKey(mv);
    }
    
    /** ONLY FOR TESTS DONT USE THEM IN ANOTHER WAY 
     * @return mapping to mv */
    Term valueOf(Metavariable mv) {
	return map.get(mv);
    }

    /** @return String representation of the constraint */
    public String toString() {
	return map.toString();
    }

    /** inner class used to store a possible mapping from a
     * Metavariable to a Term */
    class EntryMV_T implements EntryOfMetavariableAndTerm {
	private Metavariable mv;
	private Term t;

	/** creates entry of Metavariable and Term 
	 * @param mv the Metavariable mapped to Term t
	 * @param t the Term mv is mapped to
	 */
	public EntryMV_T(Metavariable mv, Term t) {
	    this.mv=mv;
	    this.t=t;
	}

	/** @return true iff content of both entry have an equal
	 * content */ 
	public boolean equals(Object obj) {
	    EntryMV_T cmp=(EntryMV_T) obj;
	    return mv.equals(cmp.key()) && t.equals(cmp.value());
	}

	/** @return the stored Metavariable */  
	public Metavariable key() {
	    return mv;
	}

	/** @return the stored Term */
	public Term value() {
	    return t;
	}
    }
}
