// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;
import de.uka.ilkd.key.util.Debug;
import java.util.Stack;
import java.util.HashSet;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
import de.uka.ilkd.key.logic.op.*;

/**
 * Simplifies updates in a term when called by
 * Term.execPostOrder(). 
 */
public class UpdateSimplifier {

    private int currentDepth=-1;
    private int maxDepth=-1;

    public UpdateSimplifier() {
    }

    /** this method simplifies the list of updates itsself	
     */
    private ListOfTerm applyOnUpdate(Term t, ListOfTerm updates) {	
	boolean changed = false;
	ListOfTerm result = SLListOfTerm.EMPTY_LIST;
	ListOfTerm processUpdates = updates;      
	UpdateOp passiveOp = (UpdateOp)t.op();
	while (processUpdates != SLListOfTerm.EMPTY_LIST) {
	    Term appliedUpdate = processUpdates.head();
	    processUpdates = processUpdates.tail();
	    
	    UpdateOp appliedOp = (UpdateOp)appliedUpdate.op();
	    if (appliedOp.variable().op() == passiveOp.variable().op()) {
		Term simplifiedUpdate=TermFactory.DEFAULT.
		    createUpdateTerm(passiveOp.variable(),
				     simplify(t.sub(0), 
					      SLListOfTerm.EMPTY_LIST.
					      prepend(appliedUpdate), 
					      true),
				     t.sub(1));
		result = result.append(simplifiedUpdate);
		passiveOp = (UpdateOp) simplifiedUpdate.op();
		changed = true;
	    } else {
		result = result.append(appliedUpdate);
		passiveOp = (UpdateOp)appliedUpdate.op();
	    }
	}
	return changed ? result : updates.prepend(t);
    }


    /** the top operator of the term t is an AttributeOperator. The
     * updates are applied on it. Starting with the first element of
     * the list.
     */
    private Term applyUpdatesOnAttribute(Term t, ListOfTerm updates) {
	Term result = t;
	AttributeOp attr = (AttributeOp)t.op();
	Term update;
	while (updates != SLListOfTerm.EMPTY_LIST) {
	    update = updates.head();
	    updates = updates.tail();

	    UpdateOp updateOp = (UpdateOp)update.op();
	    Term updateVariable = updateOp.variable();
	    if (updateVariable.op() instanceof ProgramVariable ||
		attr.attribute() != ((AttributeOp)updateVariable.op())
		.attribute()) {
		// {x:=E}o.a ----> ({x:=E}o).a or
		// {u.b:=E}o.a -----> ({u.b:=E}o).a (where a!=b) 
		Term oldResult = result;				
		result = simplify(result.sub(0), 
				  SLListOfTerm.EMPTY_LIST.prepend(update),
				  true);

		if (result != oldResult.sub(0)) {
		    result = TermFactory.DEFAULT.
			createAttributeTerm((ProgramVariable)attr.attribute(), 
					    result.sort(),
					    result);
		} else {
		    result = oldResult;
		}
		
	    } else {
		// introduce conditional term
		// {u.a:=E}o.a -----> ({u.a:=E}o ?= u).a /-> E
		Term uaEo = TermFactory.DEFAULT.
			createUpdateTerm(updateOp,
					 update.sub(0), 
					 result.sub(0));
		uaEo = simplify(uaEo);
		// if we can resolve a conditional term here we do it
 		if (uaEo.op() instanceof ProgramVariable && 
 		    uaEo.op() == updateVariable.sub(0).op()) {
 		    result = update.sub(0);
 		} else if (uaEo.equals(updateVariable.sub(0))) {
		    result = update.sub(0);
		} else {
		    result = TermFactory.DEFAULT.createCondTerm
			((ProgramVariable)attr.attribute(), uaEo.sort(), 
			 uaEo, updateVariable.sub(0), 
			 update.sub(0));
 		}
		return simplify(result, updates, true);
	    }
	    
	}
	return addUpdatesBefore(result, updates);
    }


    /** the top operator of the term t is a ProgramVariable. The
     * updates are applied on it. Starting with the first element of
     * the list.
     */
    private Term applyUpdatesOnProgramVariable(Term t, ListOfTerm updates) {
	Term term;
	while (updates != SLListOfTerm.EMPTY_LIST) {
	    term = updates.head();
	    updates = updates.tail();
	    if (((UpdateOp)term.op()).variable().op() == t.op()) {
		// if application has performed we have to descend at
		// the replaced term
		return simplify(term.sub(0), updates, true);
	    }
	}
	return t;
    }

    /** the updates in the list (the top operators of their terms and
     * their 0-th subterm (the update) are put in front of the given
     * term because (e.g) the top operator of t is a modality and
     * these updates cannot be pushed about the modality)
     */
    private Term addUpdatesBefore(Term t, ListOfTerm updates) {
	Term result = t;
	IteratorOfTerm it = updates.iterator();
	Term term;
	while (it.hasNext()) {
	    term = it.next();
	    result = TermFactory.DEFAULT.createUpdateTerm
		((UpdateOp)term.op(), term.sub(0), result);
	}
	return result;
    }

    /** call simplify on all subterm of the given term with updates handed
     * through
     */
    private Term applyOnSubTerms(Term t, ListOfTerm updates, boolean pushed) {	
	Term result = t;
	// the subterms
	Term subResult[] = new Term[t.arity()];
	ArrayOfQuantifiableVariable vars[] = new 
	    ArrayOfQuantifiableVariable[t.arity()];	    
	boolean subTermHasChanged = false;
	// application on the subterms
	for (int i = 0; i < t.arity(); i++) {
	    vars[i] = t.varsBoundHere(i);    
	    if (currentDepth>maxDepth) {
		subResult[i] = t.sub(i); //optimization
	    } else {
		if (currentDepth>0) currentDepth++;	
		subResult[i] = simplify(t.sub(i), updates, true);
		if (currentDepth>0) currentDepth--;
	    }
	    if (subResult[i] != t.sub(i)) {
		subTermHasChanged = true;
	    }
	}	    
	if (subTermHasChanged) {
	    if (t.op() instanceof CondTermOp && 
		subResult[0].op() instanceof ProgramVariable && 
		subResult[0].op() == subResult[1].op()) {
		    result = subResult[2];
	    } else {
		result = TermFactory.DEFAULT.createTerm(t.op(), subResult, 
							vars, t.javaBlock(),
							t.ruleVar());
	    }
	}	    
	return result;
    }


    /** the updates found as top operator of the list elements are
     * applied to the given term
     * @param t the Term on which the given updates are applied
     * @param updates the ListOfTerm with the updates 
     * @param pushed a boolean that marks if the updates have been
     * pushed over some other operators, so they have to be added if
     * in front of a modality(a new term has to be created) iff they
     * have not been pushed we can save memory by returning the same term
     * @return the Term after applying the updates 
     */
    private Term simplify(Term t, ListOfTerm updates, boolean pushed) {	
	Term result = t;
	// check cases where we have to do something
	if (t.op() instanceof Modality) {
	    // simplify is called on the subterms of the modal
	    // operator but starts with an empty list
	    result = simplify(t.sub(0), SLListOfTerm.EMPTY_LIST, false);
	    // we have to add updates before the modality iff the
	    // subterm of the modal term has changed or the updates
	    // have been pushed through some subterms before they
	    // arrived this modal term e.g. {x:=2}{y:=3}(A and <{}>B) 
	    // B will not be changed but the updates have to be added
	    // before <{}>B  	    
	    if (result != t.sub(0)) { 
		// subterm of modality has changed
		result = TermFactory.DEFAULT.
		    createProgramTerm((Modality)t.op(), t.javaBlock(), result);	    		
	    } else {
		result = t;
	    }
	    if (pushed || result != t.sub(0)) {
		result = addUpdatesBefore(result, updates);		
	    } 
	    return result;
	} else if (t.op() instanceof ProgramVariable) {
	    // here we have to apply the updates on the variable
	    result = applyUpdatesOnProgramVariable(t, updates); 	    
	    return result;
	} else if (t.op() instanceof UpdateOp) {
	    // we have to put the update operator in front of the list
	    // and to avoid that the 0-th subterm of this operator is
	    // updated by the list of updates
	    ListOfTerm oldUpdates = updates;
	    if (updates != SLListOfTerm.EMPTY_LIST) {
		updates = applyOnUpdate(t, updates);		
	    } else {
		updates = updates.prepend(t);
	    }
	    result = simplify(t.sub(1), updates, pushed);    	    
	    if (result.op() instanceof UpdateOp && result.sub(1) == t.sub(1)
		&& oldUpdates == updates) {
		return t;
	    } else {
		return result;
	    }
	} else if (t.op() instanceof AttributeOp) {
	    // they have to be handled special ,e.g condititonal
	    // updates have to be used
	    result = applyUpdatesOnAttribute(t, updates); 	    
	    // do not call recursive
	    return result;
	} else if (t.op() instanceof CondTermOp) {
	    // if the top operator is a conditional term operator we can only
	    // descend if the attribute of the first update in list is not the
	    // same like the conditional term operator otherwise we have to
	    // add the updates before
	    ListOfTerm updatesThatCanBePushedThrough = SLListOfTerm.EMPTY_LIST; 
	    Term update;
	    while (updates != SLListOfTerm.EMPTY_LIST) {
		update = updates.head();
		Term updateVariableTerm = ((UpdateOp)update.op()).variable();
		ProgramVariable updateVariable;
		if (updateVariableTerm.op() instanceof ProgramVariable) {
		    updateVariable = (ProgramVariable)updateVariableTerm.op();
		} else {
		    updateVariable 
		       =(ProgramVariable)((AttributeOp)updateVariableTerm.op()).attribute();
		}
		if (((CondTermOp)result.op()).attribute().attribute() == updateVariable) {
		    result = applyOnSubTerms(t, updatesThatCanBePushedThrough, pushed);		    		    
		    return addUpdatesBefore(result, updates);
		} else {
		    // collect uncritical updates
		    updatesThatCanBePushedThrough = 
			updatesThatCanBePushedThrough.append(update);

		}		
		updates = updates.tail();		
	    }
	    if (updatesThatCanBePushedThrough != SLListOfTerm.EMPTY_LIST) {
		result = applyOnSubTerms(t, updatesThatCanBePushedThrough,
					 pushed);
	    } 
	}  else {
	    result = applyOnSubTerms(t, updates, pushed);
	}
	return result;	
    }
    

    /** the updates found are pushed through and
     * applied on the term as far as possible
     * @param t the Term on which the given updates are applied
     * @return the Term after applying the updates 
     */
    public Term simplify(Term t) {
	return deleteUselessUpdates
	    (simplify(t, SLListOfTerm.EMPTY_LIST, false));
    }

    public Term simplify(Term t, int maxDepth) {
	this.maxDepth=maxDepth;
	currentDepth=0;
	return simplify(t);
    }

    private ResultWrapper collectRightHandVariables(Term t) {
	Term term = t;	
	if (t.arity() == 0) {
	    HashSet result = new HashSet();
	    if (t.op() instanceof ProgramVariable) {
		result.add(t.op());
	    }
	    return new ResultWrapper(result, term);
	}
			
	ResultWrapper[] vars = new ResultWrapper[t.arity()];	

	HashSet diamondVars = new HashSet();
	if (t.javaBlock() != JavaBlock.EMPTY_JAVABLOCK) {
	    ProgramVariableCollector pvc = new ProgramVariableCollector
		(t.javaBlock().program());
	    pvc.start();
	    diamondVars = pvc.result();
	}

	for (int i = 0; i < t.arity(); i++) {
	    vars[i] = collectRightHandVariables(t.sub(i));	    
	    if (t.op() instanceof AttributeOp) {
		vars[i].set().add(((AttributeOp)t.op()).attribute());
	    }
	}	

	if (t.op() instanceof UpdateOp) {
	    Operator op = ((UpdateOp)t.op()).
		variable().op();
	    if (!(vars[1].contains(op) || vars[0].contains(op) || 
		  diamondVars.contains(op)) && 
		op instanceof ProgramVariable) {
		vars[1].set().addAll(diamondVars);
		Debug.out("updatesimplifier: delete",op);
		return new ResultWrapper(vars[1].set(), 
					 t.sub(1));
	    }
	} 
	boolean changed = false;
	ArrayOfQuantifiableVariable[] boundVars = 
	    new ArrayOfQuantifiableVariable[t.arity()];
	Term[] subs = new Term[t.arity()];
	for (int i = 0; i < t.arity(); i++) {
	    subs[i] = vars[i].term();
	    boundVars[i] = t.varsBoundHere(i);
	    if (subs[i] != t.sub(i)) {
		changed = true;
	    }
	}
	if (changed) {
	    term = TermFactory.DEFAULT.createTerm
		(t.op(), subs, boundVars, t.javaBlock(),
		 t.ruleVar());	  
	}  
	HashSet union = union(vars);
	union.addAll(diamondVars);
	return new ResultWrapper(union, term);
    }
    

    /** delete updates whose updated variable var is a programvariable(not a field
     * or array and var does not occur on any right side of an update or in the
     * formula itself
     */
    private Term deleteUselessUpdates(Term t) {
	return collectRightHandVariables(t).term();
    }
	

    private HashSet union(ResultWrapper[] sets) {
	HashSet result = new HashSet();
	for (int i = 0; i<sets.length; i++) {
	    result.addAll(sets[i].set());
	}
	return result;
    }

    public static int depthWithoutUpdates(Term t) {
	if (t.arity()==0) return 0;
	if (t.op() instanceof UpdateOp) {
	    return depthWithoutUpdates(t.sub(1));
	}
	int m=0;
	for (int i=0; i<t.arity(); i++) {	   
	    int k=depthWithoutUpdates(t.sub(i))+1;
	    if (k>m) {
		m=k;
	    }
	}
	return m;
    }


    private static class ResultWrapper {
	private HashSet set;
	private Term term;

	public ResultWrapper(HashSet set, Term term) {
	    this.set = set;
	    this.term = term;
	}

	public boolean contains(Operator op) {
	    return set.contains(op);
	}

	public Term term() {
	    return term;
	} 

	public HashSet set() {
	    return set;
	}
	
    }


}
