// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.reference.FieldReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramInLogic;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;

/** this class is used to represent the reference operator '.' of
 * Java, e.g.>> int[] a = new int[]{1,2,3};  int i = a.length; <<<
 * where the '.' in a.length points to field called length
 * ATTENTION: At this moment only variables are adressed to this may
 * change in future
 */
public class AttributeOp implements Operator, ProgramInLogic {

    /** the field this operator references to */
    private final TermSymbol attribute;

    /** the sort of class where the field is declared */
    private final Sort prefixSort;

    /** name of this operator */
    protected final Name name;

    /** creates a new attribute operator to the field the
     * FieldReference parameter adresses to
     * @param attribute the ProgramVariable this attribute 
     * operator should reference to
     * @param prefixSort the Sort of the class where the variable is
     * declared 
     */
    public AttributeOp(ProgramVariable attribute, 
		       Sort prefixSort) {
	this.attribute  = attribute;
	this.prefixSort = prefixSort;	
	this.name       = new Name("."+attribute.name());
    }

    /** creates a new attribute operator to the field the
     * FieldReference parameter adresses to
     * @param attribute the ProgramVariable this attribute 
     * operator should reference to
     * @param prefixSort the Sort of the class where the variable is
     * declared 
     */
    public AttributeOp(SchemaVariable attribute) {
	this.attribute  = attribute;
	this.prefixSort = null;	
	this.name       = new Name("."+attribute.name());
    }

    /**
     * checks whether the top level structure of the given {@link Term}
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	if (attribute instanceof SchemaVariable) {
	    return true;
	}
	if (term.arity() != arity()) {
	    return false;
	}
	if (!(term.sub(0).sort() instanceof ObjectSort)
	    && !(term.sub(0).sort()==Sort.TYPE)) {
	    return false;
	}
	if (!term.sub(0).sort().extendsTrans(prefixSort)) {
	    return false;	    
	}
	return true;
    }

    /** returns the attribute of the operator 
     * @return  the attribute of the operator 
     */
    public TermSymbol attribute() {
	return attribute;
    }
    
    public KeYJavaType getKeYJavaType() {
	return ((ProgramVariable)attribute()).getKeYJavaType();
    }

    public Expression convertToProgram(Term t, de.uka.ilkd.key.util.ExtList l) {
	return new FieldReference((ProgramVariable)attribute(), (ReferencePrefix)l.get(0));
    }

    /** returns the sort of the class where this attribute is declared
     * @return  the sort of the class where this attribute is declared
     */
    public Sort prefixSort() {
	return prefixSort;
    }

    /**
     * determines the sort of a {@link Term} with this Operator as top
     * operator. The assumption that this Operator and the top level
     * operator are equal is NOT checked.  
     * @return the sort of the
     * given Term assumed the top level operator correspnds to this
     * operator.
     */
    public Sort sort(Term term) {
	return attribute().sort();
    }

    /** returns the name of the operator 
     * @return the name of the operator */
    public Name name() {
	return name;
    }


    /** @return arity of the Operator as int */
    public int arity() {
	return 1;
    }

    /** toString */
    public String toString() {
	return "."+attribute();
    }

    public boolean equals(Object o) {
	if (!(o instanceof AttributeOp)) {
	    return false;
	}
	AttributeOp attr = (AttributeOp)o;
	return ((attr.attribute() == attribute()) &&
		(attr.prefixSort() == prefixSort()));
    }
    
} 
