// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.ocl.ocldltranslator;

/**
 * The transformation of an OCL constraint in a representation
 * in an aribrary logical system.
 */

import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;
import de.uka.ilkd.key.ocl.ocldltranslator.transform.*;

public abstract class BasicOclConstraintTranslator extends DepthFirstAdapter {
    
    protected int constraintKind;
    protected String constrainedType;
    protected String constrainedOperation;

    protected Vector predefinedVarNames;
    protected Vector predefinedVarTypes;
    protected Vector methodParameterNames;
    protected Vector methodParameterTypes;
    protected String methodReturnType;

    public BasicOclConstraintTranslator(){
	predefinedVarNames = new Vector();
	predefinedVarTypes = new Vector();	
	methodParameterNames = new Vector();
	methodParameterTypes = new Vector();
	methodReturnType = null; // no return type was specified ...
    }
    

    public String constrainedOperation() {
	return constrainedOperation;
    }
    

    // methods to separate constraint into ocl expressions 
    // for each constraint_body

    public void inAClassifierContext(AClassifierContext cc) {
	constrainedType=cc.getTypeName().toString().trim();
	
	predefinedVarNames.add("self");
	predefinedVarTypes.add(constrainedType);
    }
    
    public void inAOperationContext(AOperationContext oc) {
	constrainedType=oc.getTypeName().toString().trim();

	predefinedVarNames.add("self");
	predefinedVarTypes.add(constrainedType);

	constrainedOperation=
	    oc.getName().toString().trim();

	// If there are formal parameters then add their signature to the predefined variables ...
	if (oc.getFormalParameterList() != null) {
	    AFormalParameterList fpl=(AFormalParameterList)oc.getFormalParameterList();
	    AFormalParameter fp = (AFormalParameter)fpl.getFormalParameter();
	    
	    String paramName = fp.getName().toString().trim();
	    String paramType = fp.getTypeName().toString().trim();
	    
	    predefinedVarNames.add(paramName);
	    predefinedVarTypes.add(paramType);
	    
	    methodParameterNames.add(paramName);
	    methodParameterTypes.add(paramType);
	    
	    Iterator paramIter=fpl.getFormalParameterListTail().iterator();
	    while (paramIter.hasNext()) {
		fp = (AFormalParameter) ((AFormalParameterListTail) paramIter.next()).getFormalParameter();
		paramName = fp.getName().toString().trim();
		paramType = fp.getTypeName().toString().trim();
		
		predefinedVarNames.add(paramName);
		predefinedVarTypes.add(paramType);
		
		methodParameterNames.add(paramName);
		methodParameterTypes.add(paramType);
	    }
	    
	    
	    
	    
	}
	
	// determine result type of method ...
	if (oc.getReturnTypeDeclaration() != null) {
	    // we have a delaracrion of the return type of the method ...
	    AReturnTypeDeclaration rtd =(AReturnTypeDeclaration)oc.getReturnTypeDeclaration();
	    methodReturnType = rtd.getTypeName().toString().trim();
	
	    log("RETURN TYPE OF METHOD == " + methodReturnType);
	    // HACK : methodReturnType = "Integer";
	    predefinedVarNames.add("result");
	    predefinedVarTypes.add(methodReturnType);
	}
		
    }
    
    public final void caseAConstraintBody(AConstraintBody cb) {
	String stereoType = cb.getStereotype().toString().trim();
	String exprName= ( (cb.getName() != null) ? (cb.getName().toString().trim())
	    : ("UNNAMED_"+ stereoType));
	AExpression oclExpression = (AExpression) cb.getExpression();
	translateOclExpression(cb,oclExpression, stereoType, exprName, predefinedVarNames
			       , predefinedVarTypes);
    }
    
    
    // ================= Some helper methods ============================================
    
    protected void log(String msg){
	System.err.println(msg);
    }


     // ================== Some methods for errors ================================
    
    // a method for constructing error messages ...
    void error(String msg) throws TranslationException {
	throw (new TranslationException(msg));
    }


    // ================ Start the translation process (used by subclasses) =============
    protected void startTranslationOf(OclTree oclTree){
	oclTree.apply(this);
    }

    
    // ================= Some method to translate a complete OCL constraint =============

    public abstract TranslationResult getTranslationFor(OclTree oclTree);

    // translate a given OCL-expression of a OCL constraint body ...
    public abstract TranslationResult translateOclExpression(Node cb, AExpression oclExpression,
							     String stereoType,
							     String exprName, 
							     Vector predefinedVarNames,
							     Vector predefinedVarTypes);



}





