// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import junit.framework.*;
import java.io.*;
import de.uka.ilkd.key.parser.schemajava.ParseException;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.declaration.VariableSpecification;
import de.uka.ilkd.key.logic.op.TermSymbol;
import de.uka.ilkd.key.rule.*;


/** class tests the parser for Taclets
*/


public class TestTacletParser extends TestCase {


    public Namespace var_ns;
    public Namespace func_ns;
    public Namespace sort_ns;
    public Namespace heur_ns;
    public JavaInfo javaInfo;

    TermFactory tf=TermFactory.DEFAULT;

    public TestTacletParser(String name) {
	super(name);
    }

    //
    //  set up
    //

    public void setUp() {
	var_ns =new Namespace();
	func_ns=new Namespace();
	sort_ns=new Namespace();
	heur_ns=new Namespace();
	javaInfo = TacletForTests.javaInfo();

	parseDecls("sorts { s; }\n" +
		   "functions {\n" +
		   "  s f(s);\n" +
		   "}\n" +
		   "schema variables {\n" +
		   "  formula b,b0,post;\n" +
		   "  program statement #p1, #s; \n" +
		   "  program expression #e2, #e; \n" +
		   "  program simple expression #se; \n" +
		   "  program simple lefthandside #slhs; \n" +
		   "  program forinit #i; \n"+
		   "  program label #lab, #lb0, #lb1; \n"+
		   "  program type #typ; \n"+
		   "  program variable #v0; \n"+
		   "  program booleanvariable #boolv; \n"+
		   "  program catchfinally #cf; \n"+
		   "  program methodreference #mr; \n"+
		   "  s x,x0;\n" +		
		   "  s variables z,z0;\n" +		   
		   "}\n"
		   );
    }

    //
    // Utility methods for setUp:
    //
    
    TermSymbol lookup_var(String name) {
	return (TermSymbol)var_ns.lookup(new Name(name));
    }
    

    private DeclParser stringDeclParser(String s) {
	return new DeclParser(new KeYLexer(new StringReader(s)),
			      "No file. parser/TestTacletParser.stringDeclParser("+s+")",  
			       var_ns, func_ns, sort_ns, heur_ns);
    }

    public void parseDecls(String s) {
	try {
	    DeclParser p = stringDeclParser(s);
	    p.decls();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    //
    // Utility Methods for test cases.
    //

    private TacletParser stringTacletParser(String s) {
	return new TacletParser(new KeYLexer(new StringReader(s)), 
			      "No file. parser/TestTacletParser.stringTacletParser("+s+")",  
			      tf, javaInfo, var_ns, func_ns, sort_ns, heur_ns);
    }

    public Term parseTerm(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    return p.term();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }


    public Term parseFma(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    
	    return p.formula();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    public ConstrainedFormula cf(String s) {
	return new ConstrainedFormula(parseFma(s),Constraint.BOTTOM);
    }
    
    public Semisequent sseq(String s) {
	return Semisequent.EMPTY_SEMISEQUENT.insertFirst(cf(s));
    }
    
    public Sequent sequent(String a,String s) {
	Semisequent ass = Semisequent.EMPTY_SEMISEQUENT;
	Semisequent sss = Semisequent.EMPTY_SEMISEQUENT;
	if ( a != null ) {
	    ass = sseq(a);
	}
	if ( s != null ) {
	    sss = sseq(s);
	}
	return Sequent.createSequent(ass,sss);
    }
    
    Taclet parseTaclet(String s) {
	try {
	    TacletParser p = stringTacletParser(s);
	    
	    return p.taclet();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    //
    // Test cases.
    //
 
    public void testImpLeft() {
	// imp-left rule
	// find(b->b0 =>) replacewith(b0 =>) replacewith(=> b) 
	AntecTacletBuilder builder=new AntecTacletBuilder();
	builder.setFind(parseFma("b->b0"));
	builder.setName(new Name("imp_left"));
	builder.addTacletGoalTemplate(new
	    AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				      SLListOfTaclet.EMPTY_LIST,
				      SLListOfRulevariable.EMPTY_LIST,
				      sequent("b0",null)));
	
	builder.addTacletGoalTemplate(new
	    AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				      SLListOfTaclet.EMPTY_LIST,
				      SLListOfRulevariable.EMPTY_LIST,
				      sequent(null,"b")));
	
	Taclet impleft=builder.getAntecTaclet();
	String impleftString=
	    "imp_left{find(b->b0 ==>) replacewith(b0 ==>); replacewith(==> b)}";
   	assertEquals("imp-left",impleft,parseTaclet(impleftString));
    }
    
    public void testImpRight() {
	// imp-right rule
	// find(=> b->b0) replacewith(b => b0) 
	SuccTacletBuilder builder=new SuccTacletBuilder();
	builder.setFind(parseFma("b->b0"));
	builder.addTacletGoalTemplate(new
	    AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				      SLListOfTaclet.EMPTY_LIST,
				      SLListOfRulevariable.EMPTY_LIST,
				      sequent("b","b0")));
	builder.setName(new Name("imp_right"));
	Taclet impright=builder.getSuccTaclet();
	String imprightString=
	    "imp_right{find(==> b->b0) replacewith(b ==> b0)}";
	
 	assertEquals("imp-right",impright,parseTaclet(imprightString));
    }
    
    public void testCut() {
	// cut rule
	// add(b=>) add(=>b)
	NoFindTacletBuilder builder=new NoFindTacletBuilder();
	builder.addTacletGoalTemplate(new
	    TacletGoalTemplate(sequent("b",null),
			     SLListOfTaclet.EMPTY_LIST,
			     SLListOfRulevariable.EMPTY_LIST
			     ));
	
	builder.addTacletGoalTemplate(new
	    TacletGoalTemplate(sequent(null,"b"),
			     SLListOfTaclet.EMPTY_LIST,
			     SLListOfRulevariable.EMPTY_LIST
			     ));
	builder.setName(new Name("cut"));
	
       	Taclet cut=builder.getNoFindTaclet();
	String cutString = "cut{add(b==>); add(==>b)}";
	assertEquals("cut",cut,parseTaclet(cutString));
    }

    public void testClose() {
	// close rule
	// if (b=>) find(=>b) 
	SuccTacletBuilder builder=new SuccTacletBuilder();
	builder.setFind(parseFma("b"));
	builder.setIfseq(sequent("b",null));
	builder.setName(new Name("close_goal"));
       	Taclet close=builder.getSuccTaclet();
	String closeString = "close_goal{if (b==>) find(==>b) close goal}";
	assertEquals("close",close,
		     parseTaclet(closeString));
    }

    public void testContraposition() {
	// contraposition rule
	// find(b->b0) replacewith(-b0 -> -b)

	RewriteTacletBuilder builder=new RewriteTacletBuilder();
	builder.setFind(parseFma("b->b0"));
	builder.addTacletGoalTemplate(new
	    RewriteTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				    SLListOfTaclet.EMPTY_LIST,
				    SLListOfRulevariable.EMPTY_LIST,
				    parseFma("!b0->!b")));
	builder.setName(new Name("contraposition"));
       	Taclet contraposition=builder.getRewriteTaclet();
	String contrapositionString=
	    "contraposition{find(b->b0) replacewith(!b0 -> !b)}";

	assertEquals("contraposition",contraposition,
		     parseTaclet(contrapositionString));
    }

    public void testAllRight() {
	// all-right rule
	// find(==>all z . b) replacewith(==>{z  eps z.!b}b)
 	SuccTacletBuilder builder=new SuccTacletBuilder();
	
	builder.setFind(parseFma("all z.b"));
	builder.addTacletGoalTemplate(new
	    AntecSuccTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				      SLListOfTaclet.EMPTY_LIST,
				      SLListOfRulevariable.EMPTY_LIST,
				      sequent(null,"{z eps z.!b}b")));
	builder.setName(new Name("all_right"));
       	Taclet allright=builder.getSuccTaclet();
	String allrightString=
	    "all_right{find(==>all z . b) replacewith(==>{z eps z.!b}b)}";
	assertEquals("all-right",allright,
		     parseTaclet(allrightString));
    }
    
    public void testAllLeft() {
	// all-left rule
	// find(all z . b==>) add({z x}b==>)
 	AntecTacletBuilder builder=new AntecTacletBuilder();

	builder.setFind(parseFma("all z.b"));
			       
	
	builder.addTacletGoalTemplate(new
	    TacletGoalTemplate(sequent("{z x}b",null),
			     SLListOfTaclet.EMPTY_LIST,
			     SLListOfRulevariable.EMPTY_LIST));
	builder.setName(new Name("all_left"));
       	Taclet allleft=builder.getAntecTaclet();
	String allleftString = 
	    "all_left{find(all z . b==>) add({z x}b==>)}";
	assertEquals("all-left",allleft,
		     parseTaclet(allleftString));
    }

    public void testExConjSplit() {
 	//exists-conj-split rule
	//find(ex z . ( b & b0 )) varcond(z not free in b) 
	//  replacewith( b & ex z.b0 )
	
	RewriteTacletBuilder builder=new RewriteTacletBuilder();
	builder.setFind(parseFma("ex z.(b & b0)"));
	builder.addVarsNotFreeIn((SchemaVariable)lookup_var("z"),
				   (SchemaVariable)lookup_var("b"));
	builder.addTacletGoalTemplate(new
	    RewriteTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				    SLListOfTaclet.EMPTY_LIST,
				    SLListOfRulevariable.EMPTY_LIST,
				    parseFma("b & ex z.b0")));
	builder.setName(new Name("exists_conj_split"));
	Taclet exconjsplit = builder.getRewriteTaclet();
	String exconjsplitString = "exists_conj_split{"
	    + "find(ex z . ( b & b0 )) varcond(z not free in b) \n"
	    +"replacewith( b & ex z.b0 )}";
	assertEquals("ex-conj-split",exconjsplit,
		     parseTaclet(exconjsplitString));
    }

    public void testFIdempotent() {
 	// f-idempotent-rule
	// find(f(f(x))) replacewith( f(x) )
	
	RewriteTacletBuilder builder=new RewriteTacletBuilder();

	builder.setFind(parseTerm("f(f(x))"));
	builder.addTacletGoalTemplate(new
	    RewriteTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				    SLListOfTaclet.EMPTY_LIST,
				    SLListOfRulevariable.EMPTY_LIST,
				    parseTerm("f(x)")));
	builder.setName(new Name("f_idempotent"));
	Taclet fidempotent = builder.getRewriteTaclet();
	String fidempotentString = 
	    "f_idempotent{find(f(f(x))) replacewith(f(x))}";
	assertEquals("f-idempotent",fidempotent,
		     parseTaclet(fidempotentString));
    }

    public void testMakeInsertEq() {
	// make-insert-eq rule
	// find (x = x0 =>) addrules ( find (x) replacewith (x0) )
	RewriteTacletBuilder insertbuilder=new RewriteTacletBuilder();
	insertbuilder.setFind(parseTerm("x"));
	insertbuilder.addTacletGoalTemplate(new
	    RewriteTacletGoalTemplate(Sequent.EMPTY_SEQUENT,
				    SLListOfTaclet.EMPTY_LIST,
				    SLListOfRulevariable.EMPTY_LIST,
				    parseTerm("x0")));
	insertbuilder.setName(new Name("insert_eq"));
	Taclet inserteq = insertbuilder.getTaclet();
	
	AntecTacletBuilder builder = new AntecTacletBuilder();
	
	builder.setFind(parseFma("x=x0"));
	builder.addTacletGoalTemplate(new
	    TacletGoalTemplate(Sequent.EMPTY_SEQUENT,
			     SLListOfTaclet.EMPTY_LIST.prepend(inserteq),
			     SLListOfRulevariable.EMPTY_LIST));
	builder.setName(new Name("make_insert_eq"));
	Taclet makeinserteq = builder.getAntecTaclet();
	String makeinserteqString = "make_insert_eq"
	    +"{find (x = x0 ==>)"
	    + "addrules ( insert_eq{find (x) replacewith (x0)} )}";
	assertEquals("make-insert-eq",makeinserteq,
		     parseTaclet(makeinserteqString));
    }

    public void testSchemaJava0()  {
	
	    parseTaclet("while_right { find (<{.. while(#e2) {#p1} ...}>post)"
	    +"replacewith (<{.. #unwind-loop (while(#e2) {#p1});  ...}>post) } ");

    }

    public void testSchemaJava4() {
	FindTaclet taclet=	(FindTaclet) parseTaclet("variable_declaration{ find (<{.. #typ #v0; ...}>post)"
		  +" replacewith (<{.. #typ #v0; if (true); ...}>post)	}");
	Term find=taclet.find();
	JavaBlock jb=find.javaBlock();

	ContextTerm ct=(ContextTerm)jb.program();
	StatementBlock sb=(StatementBlock)ct.getChildAt(0);
	LocalVariableDeclaration lvd=(LocalVariableDeclaration)sb.getChildAt(0);
	VariableSpecification vs=(VariableSpecification)lvd.getChildAt(1);

    }

    public void testSchemaJava5() {
	RewriteTaclet taclet=	(RewriteTaclet) 
	    parseTaclet("xy{ find (<{.. try{ throw #e; #s} #cf ...}>post)"
		      +" replacewith (<{.. #try-throw-to-if(try{throw #e; #s} #cf); ...}>post)	}");
	
	Term rw=((RewriteTacletGoalTemplate)taclet.goalTemplates().head()).replaceWith();
	JavaBlock jb=rw.javaBlock();
	ContextTerm ct=(ContextTerm)jb.program();
	taclet=	(RewriteTaclet) 
	    parseTaclet("xy{ find (<{.. try{ } #cf ...}>post)"
		      +" replacewith (<{.. #empty-try(try{ } #cf); ...}>post)	}");
	
	rw=((RewriteTacletGoalTemplate)taclet.goalTemplates().head()).replaceWith();
	jb=rw.javaBlock();
	ct=(ContextTerm)jb.program();

    }

    public void testVarcondNew() {
	FindTaclet taclet = (FindTaclet) parseTaclet("xy{ find (true) varcond (s #boolv new) replacewith(true)}");

	taclet = (FindTaclet) parseTaclet("xy{ find (true) varcond (typeof(#e2) #v0 new) replacewith(true)}");

    }

    public void testSchemaJava6() {
	FindTaclet taclet=	(FindTaclet) parseTaclet("xy{ find (<{.. boolean #boolv; ...}>post)"
		  +" replacewith (<{.. if (true); ...}>post)	}");
	Term find=taclet.find();
	JavaBlock jb=find.javaBlock();

	ContextTerm ct=(ContextTerm)jb.program();
	StatementBlock sb=(StatementBlock)ct.getChildAt(0);
	LocalVariableDeclaration lvd=(LocalVariableDeclaration)sb.getChildAt(0);
	VariableSpecification vs=(VariableSpecification)lvd.getChildAt(1);
    }

    public void testSchemaJava7() {
	FindTaclet taclet=	(FindTaclet) parseTaclet
	    (" instanceof_evaluate  { find (<{.. #slhs = #e instanceof #typ; ...}>post) "
	     +"replacewith (<{.. #slhs = #static-evaluate(#e instanceof #typ);  ...}>post)}"); 
	Term find=taclet.find();
	JavaBlock jb=find.javaBlock();
	ContextTerm ct=(ContextTerm)jb.program();
    }

    public void testSchemaJava9() {
	FindTaclet taclet=	(FindTaclet) parseTaclet
	    (" testmethodcall  { find (<{.. #mr ...}>post) "
	     +"replacewith (<{.. #method-call(#mr); ...}>post)}"); 
	Term find=taclet.find();
	JavaBlock jb=find.javaBlock();
	ContextTerm ct=(ContextTerm)jb.program();
    }


    public void testSchemaJava8() {
	FindTaclet taclet=	(FindTaclet) parseTaclet
	    ("break_test {find(<{.. #lb0:{ break #lb1; } ...}>post)"+ 
	     " replacewith (<{..  ...}>post)}"); 
	Term find=taclet.find();
	JavaBlock jb=find.javaBlock();
	ContextTerm ct=(ContextTerm)jb.program();
    }


    //Following tests should work, if parser didn't exit system but propagated 
    // exceptions until here.

    /*    public void testSchemaJava1()  {
	boolean thrown=false;
	try {
	    parseTaclet("xyz { find (<{.. int j=0; for(#i;j<9;j++)"+
	    "; ...}>post)"
		      +"replacewith (<{.. #unwind-loop (while(#e2) {#p1});"+
		      "  ...}>post) } ");
	} catch (RuntimeException e) {
	    thrown=true;
	}
	assert("forinit only valid in initializer of for", thrown);
	}


    public void testSchemaJava2()  {
	boolean thrown=false;
	try {
	    parseTaclet("xyz { find (<{.. int j=0; for(#lab;j<42;j++) ; ...}>post)"
		      +"replacewith (<{.. #unwind-loop (while(#e2) {#p1});  ...}>post) } ");
	} catch (RuntimeException e) {
	    thrown=true;
	}
	assert("label SV not valid here", thrown);
    }
    */
    //more SchemaJava tests are in ../rule/TestMatchTaclet

}
