// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser;

import junit.framework.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.java.*;
import java.io.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.rule.TacletForTests;


public class TestTermParser extends TestCase {
    
    private TermFactory tf = TermFactory.DEFAULT;

    Namespace sort_ns, func_ns, var_ns, heur_ns;

    Sort elem,list;

    Function head,tail,nil,cons,isempty; 

    LogicVariable x,y,z,xs,ys,zs;

    Term t_x,t_y,t_z,t_xs,t_ys,t_zs;
    Term t_headxs,t_tailxs,t_tailys,t_nil;

    public TestTermParser(String name) {
	super(name);
    }


    public void setUp() {
	sort_ns = new Namespace();
	func_ns = new Namespace();
	var_ns = new Namespace();
	heur_ns = new Namespace();

	parseDecls("sorts { elem; list; }\n" +
		   "functions {\n" +
		   "  elem head(list);\n" +
		   "  list tail(list);\n" +
		   "  list nil;\n" +
		   "  list cons(elem,list);\n"  +
		   "}\n" +
		   "predicates {\n" +
		   "  isempty(list);\n" +
		   "}\n"
		   );
	
	elem = lookup_sort("elem");
	list = lookup_sort("list");	

	head = lookup_func("head"); tail = lookup_func("tail");
	nil = lookup_func("nil"); cons = lookup_func("cons");
	isempty = lookup_func("isempty");
	
	// The declaration parser cannot parse LogicVariables; these
	// are normally declared in quantifiers, so we introduce them
	// ourselves!
	x = declareVar("x",elem);   t_x = tf.createVariableTerm(x);  
	y = declareVar("y",elem);   t_y = tf.createVariableTerm(y);  
	z = declareVar("z",elem);   t_z = tf.createVariableTerm(z);  
	xs = declareVar("xs",list); t_xs = tf.createVariableTerm(xs);
	ys = declareVar("ys",list); t_ys = tf.createVariableTerm(ys);
	zs = declareVar("zs",list); t_zs = tf.createVariableTerm(zs);
	
	t_headxs = tf.createFunctionTerm(head,t_xs);
	t_tailxs = tf.createFunctionTerm(tail,t_xs);
	t_tailys = tf.createFunctionTerm(tail,t_ys);
	t_nil = tf.createFunctionTerm(nil);
    }

    Sort lookup_sort(String name) {
	return (Sort)sort_ns.lookup(new Name(name));
    }
    
    Function lookup_func(String name) {
	return (Function)func_ns.lookup(new Name(name));
    }

    LogicVariable declareVar(String name,Sort sort) {
	LogicVariable v = new LogicVariable(new Name(name),sort);
	var_ns.add(v);
	return v;
    }
    

    private DeclParser stringDeclParser(String s) {
	return new DeclParser(new KeYLexer(new StringReader(s)),
			      "No file. Call of parser from parser/TestTermParser.java",
			      var_ns, func_ns, sort_ns, heur_ns);
    }

    public void parseDecls(String s) {
	try {
	    DeclParser p = stringDeclParser(s);
	    p.decls();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    private TermParser stringTermParser(String s) {
	return new TermParser(new KeYLexer(new StringReader(s)), 
			      "No file. Call of parser from parser/TestTermParser.java",
			      tf, TacletForTests.javaInfo(), var_ns, func_ns, sort_ns);
    }

    public Term parseTerm(String s) {
	try {
	    TermParser p = stringTermParser(s);
	    return p.term();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }


    public Term parseFma(String s) {
	try {
	    TermParser p = stringTermParser(s);
	    
	    return p.formula();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    public void test1() {
	assertEquals("parse x",t_x,parseTerm("x"));
    }

    public void test2() {
	assertEquals("parse head(xs)",t_headxs,parseTerm("head(xs)"));
    }

    public void test3() {
	Term t = tf.createFunctionTerm(cons,t_headxs,t_tailys);
	
	assertEquals("parse cons(head(xs),tail(ys))",
		     t,parseTerm("cons(head(xs),tail(ys))"));
    }

    public void test5() {
	Term t = tf.createEqualityTerm
	    (tf.createFunctionTerm
	     (head,
	      tf.createFunctionTerm(cons, t_x, t_xs)),
	     tf.createFunctionTerm
	     (head,
	      tf.createFunctionTerm (cons,t_x,t_nil)));
	     
	assertEquals("parse head(cons(x,xs))=head(cons(x,nil))",
		     t,parseFma("head(cons(x,xs))=head(cons(x,nil))"));
    }

    public void test6() {
	Term t = tf.createJunctorTerm
	    (Op.EQV,
	     tf.createJunctorTerm
	     (Op.IMP,
	      tf.createJunctorTerm
	      (Op.OR,
	       tf.createEqualityTerm(t_x,t_x),
	       tf.createEqualityTerm(t_y,t_y)),
	      tf.createJunctorTerm
	      (Op.AND,
	       tf.createEqualityTerm(t_z,t_z),
	       tf.createEqualityTerm(t_xs,t_xs))),
	     tf.createJunctorTerm
	     (Op.NOT,
	      tf.createEqualityTerm(t_ys,t_ys)));

	     
	assertEquals("parse x=x | y=y -> z=z & xs =xs <-> ! ys = ys",
		     t,parseFma("x=x|y=y->z=z&xs=xs<->!ys=ys"));
    }

    public void test7() {
	/** Bound variables are newly created by the parser,
	 * so we have to parse first, then extract the used variables,
	 * then build the formulae. */
	
	String s = "all x:list. all l1:list. ! x = l1";
	Term t = parseFma(s);
	
	LogicVariable thisx = (LogicVariable) t.varsBoundHere(0)
	    .getQuantifiableVariable(0);
	LogicVariable l1 = (LogicVariable) t.sub(0).varsBoundHere(0)
	    .getQuantifiableVariable(0);

	Term t1 = tf.createQuantifierTerm
	    (Op.ALL,thisx,
	     tf.createQuantifierTerm
	     (Op.ALL,l1,
	      tf.createJunctorTerm
	      (Op.NOT,
	       tf.createEqualityTerm(tf.createVariableTerm(thisx),
				     tf.createVariableTerm(l1)))));
	
	assert("new variable in quantifier", thisx != x);
	assertEquals("parse all x:list. all l1:list. ! x = l1", t1,t);
	
    }

    public void test8() {
	/** A bit like test7, but for a substitution term */

	String s = "{xs:elem head(xs)} cons(xs,ys)";
	Term t = parseTerm(s);

	LogicVariable thisxs = (LogicVariable) t.varsBoundHere(1)
	    .getQuantifiableVariable(0);
	
	Term t1 = tf.createSubstitutionTerm
	    (Op.SUBST,
	     thisxs, t_headxs,
	     tf.createFunctionTerm
	     (cons, 
	      tf.createVariableTerm(thisxs), 
	      t_ys));

	assert("new variable in subst term", thisxs != xs);
	assertEquals("parse {xs:elem head(xs)} cons(xs,ys)",t1,t);
    }

    public void test9() {
	/* Try something with a prediate */
	
	String s = "ex x:list. !isempty(x)";
	Term t = parseFma(s);
	
	LogicVariable thisx = (LogicVariable) t.varsBoundHere(0)
	    .getQuantifiableVariable(0);

	Term t1 = tf.createQuantifierTerm
	    (Op.EX,thisx,
	     tf.createJunctorTerm
	     (Op.NOT,
	      tf.createFunctionTerm(isempty,tf.createVariableTerm(thisx))));
	      
	assert("new variable in quantifier", thisx != x);
	assertEquals("parse all x:list. all l1:list. ! x = l1", t1,t);
	
    }

    public void test10() {
	// Unquoted, this is
	// <{ int x = 2; {String x = "\"}";} }> true
	//	String s = "< { int x = 1; {String s = \"\\\"}\";} } > true";
	String s = "< { int x = 1; {int s = 2;} } > x=x";
	Term t = parseTerm(s);
	
	// for now, just check that the parser doesn't crash
	
	//	 System.out.println(t);

	// Same with a box
	s = "[{ int x = 2; {String s = \"\\\"}\";} }] true";
	//s = "< { int x = 1; {int s = 2;} } > true";
	t = parseTerm(s);
	//System.out.println(t);
    }

    public void test10a() {
	// Unquoted, this is
	// <{ int x = 2; {String x = "\"}";} }> true
	//	String s = "< { int x = 1; {String s = \"\\\"}\";} } > true";
	String s = "< { int x = 1; int y=2; int z=3; int w=4;} > (x ?= y).z /->w";
	//Term t = parseTerm(s);
	
	//	System.out.println(t);
    }


    public void test11() {
	/** And finally an epsilon term.  
	    This is very important.
	    Guess who wrote these tests :-) */
	
	String s = "eps x:elem . ( xs = cons(x,ys) )";
	Term t = parseTerm(s);

	LogicVariable thisx = (LogicVariable) t.varsBoundHere(0)
	    .getQuantifiableVariable(0);
	
	Term t1 = tf.createEpsilonTerm
	    (thisx,
	     tf.createEqualityTerm
	     (t_xs,
	      tf.createFunctionTerm
	      (cons,
	       tf.createVariableTerm(thisx),
	       t_ys)));
	
	assert("new variable in epsilon term", thisx != x);
	assertEquals("parse eps x:elem . xs = cons(x,ys)",t1,t);
    }

    public void test12() {
	String s="<{int i; i=0;}> <{ while (i>0) ;}>true";
	Term t=parseTerm(s);
    }


}
