// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.*;

    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.java.JavaReader;
    import de.uka.ilkd.key.java.JavaInfo;
    import java.util.Vector;
    import java.util.HashSet;
    import java.util.Iterator;
    import java.util.HashMap;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.proof.UseInsteadMap;
    import de.uka.ilkd.key.java.Recoder2KeY;
    import de.uka.ilkd.key.java.Services;
}

/** 
 * A parser for first-order terms and formulae.
 */  

class TermParser extends DeclParser;
options {
	importVocab=KeYLexer;
	k = 2;
	defaultErrorHandler=false;
}


{

    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected Namespace var_ns;
    protected Namespace func_ns;
    protected Namespace sort_ns;
    protected Namespace heur_ns;
    protected JavaInfo javaInfo;

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;

    protected int lineOffset=0;
    protected int colOffset=0;

    protected TermFactory tf;
    protected JavaReader javaReader;

    protected UseInsteadMap useInstead=new UseInsteadMap();	
    boolean schematicJava = false;	

    public boolean errors = false;


    public TermParser(TokenStream lexer,					
		      String filename,
		      TermFactory tf,
		      JavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns) {
	this(lexer,2);
	setFilename(filename);
	this.javaReader = jr;
	this.tf = tf;
	this.var_ns = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;
    }

    public TermParser(TokenStream lexer,
		      String filename,
		      JavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns) {

	this(lexer, filename, TermFactory.DEFAULT, jr, 
	     var_ns, func_ns, sort_ns);
    }

    public TermParser(TokenStream lexer,
		      String filename,
		      TermFactory tf,
		      JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns) {
	this(lexer, filename, tf, getRecoder2KeY(javaInfo),	
	     var_ns, func_ns, sort_ns);
    }

    static Recoder2KeY getRecoder2KeY(JavaInfo javaInfo) {
	 return new de.uka.ilkd.key.java.Recoder2KeY
		(javaInfo.rec2key(), javaInfo.getTypeConverter());	
    }

    public TermParser(TokenStream lexer,
		      String filename,
		      JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns) {		

	this(lexer, filename, TermFactory.DEFAULT, getRecoder2KeY(javaInfo),
	     var_ns, func_ns, sort_ns);
    }

    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine()+lineOffset;
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int 	col = -1;
	try {
	    col = LT(0).getColumn()+colOffset;
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   

    protected LogicVariable bindVar(Token id,Sort s) {
	LogicVariable v=new LogicVariable(new Name(id.getText()), s);
	var_ns=var_ns.extended(v);
	return v;
    }

    protected TermSymbol progVar(Token id,Sort s) {
	ProgramVariable v = (ProgramVariable)
	    var_ns.lookup(new Name(id.getText()));
	if (v == null) {
	    v = new ProgramVariable(new ProgramElementName(id.getText()), s);
	    System.err.println("ProgVar "+v+" not found");
	    var_ns=var_ns.extended(v);
	}
	return v;
    }


    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	if (javaInfo==null) {
           javaInfo=javaReader.createJavaInfo(new Services());
        }
	return javaInfo;
    }

	
    protected void unbindVars() {
	var_ns = var_ns.parent();
    }

    protected void bindProgVars(HashSet progVars) {
	var_ns=new Namespace(var_ns);
	Iterator it=progVars.iterator();
	while (it.hasNext()) {
	    var_ns.add((ProgramVariable)it.next());
	}
    }

    protected void unbindProgVars() {
	var_ns = var_ns.parent();
    }

    protected static HashSet progVars(JavaBlock jb) {
	if (jb==JavaBlock.EMPTY_JAVABLOCK) {
		return new HashSet();
	}	
	ProgramVariableCollector pvc
	    = new ProgramVariableCollector(jb.program());
	pvc.start();
	return pvc.result();
    }


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{

	    if ( v instanceof LogicVariable ) {
		return tf.createVariableTerm((LogicVariable)v);
	    } else if ( v instanceof ProgramVariable ) {
		return tf.createVariableTerm((ProgramVariable)v);
	    } else {
		throw new antlr.SemanticException(
						  ""+ v +" is not a logic variable");
	    }
	}

      

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
	String s=t.getText();
	try {
		return javaReader.readBlockWithProgramVariables(var_ns, s);
	} catch (de.uka.ilkd.key.java.ConvertException e) {
		if (s.length()>10) {
			s=s.substring(0,10)+"...";
		}		
		if (e.parseException()!=null) {		  		  
		  lineOffset=e.parseException().currentToken.next.beginLine;		  		  colOffset=e.parseException().currentToken.next.beginColumn;
		  e.parseException().currentToken.next.beginLine=getLine()-1;
		  e.parseException().currentToken.next.beginColumn=getColumn();
		  throw new JavaParserException(e.parseException().getMessage(), t, 
			getFilename(), -1, -1);  // row/columns already in text
		}		
		throw new JavaParserException(e.getMessage(), t, getFilename(), 0, 1);
	}
    }

}

/**
 * In the special but important case of Taclets, we don't yet know
 * whether we are going to have a term or a formula, and it is hard
 * to decide a priori what we are looking at.  So the `term' 
 * non-terminal will recognize a term or a formula.  The `formula'
 * reads a formula/term and throws an error if it wasn't a formula.
 * This gives a rather late error message. */

top {Term t;} : t=formula;

formula returns [Term a = null] 
	:
		a = term 
		{
			if ( a.sort() != Sort.FORMULA ) {
				throw new antlr.SemanticException(
					"Just Parsed a Term where a Formula was expected");
			}
		}
	;


term returns [Term a = null] 
	:
		a=term20 
	;

term20 returns [Term a = null] {Term a1;}
	:   a=term30 
		(EQV a1=term20 
			{ a = tf.createJunctorTerm(Op.EQV,new Term[]{a,a1});} ) ?;
term30 returns [Term a = null] {Term a1;}
	:   a=term40 
		(IMP a1=term30 
			{ a = tf.createJunctorTerm(Op.IMP,new Term[]{a,a1});} ) ?;
term40 returns [Term a = null] {Term a1;}
	:   a=term50 
		(OR a1=term40 
			{ a = tf.createJunctorTerm(Op.OR, new Term[]{a,a1});} ) ?;
term50 returns [Term a = null] {Term a1;}
	:   a=term60 
		(AND a1=term50 
			{ a = tf.createJunctorTerm(Op.AND,new Term[]{a,a1});} ) ?;

term60 returns [Term a = null] 
{
    Term a1,a2;
	Sort s;
}
	: (LPAREN IDENT GUARD)=> a=condterm
	|   a = term70 
	|   a=bindingterm
	|   a=dl_term		   
    |   NOT a1=term60 { a = tf.createJunctorTerm(Op.NOT,new Term[]{a1}); }    
;


term70 returns [Term a = null] {Term a1;}
	:   a=term80 
		(EQUALS a1=term80 
			{ a = tf.createEqualityTerm(a,a1);} ) ?;


term80 returns [Term a = null] 
	:
     (IDENT DOT IDENT)=> a=attrterm
    |   a=funcpredvarterm   
    |   LPAREN a=term RPAREN
    |   "true" { a = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { a = tf.createJunctorTerm(Op.FALSE); }
    ;



condterm returns [Term a = null]
{ Term a1=null;
  Term a2=null;
  Term a3=null;
  ParsableVariable v=null;
}
       : 
    LPAREN a1=term80 GUARD a2=term80 RPAREN DOT v=var_id CONDUP a3=term60 {
	AttributeOp aop=null;
	aop=new AttributeOp((ProgramVariable)v,a1.sort());	
	a = tf.createCondTerm(aop,a1,a2,a3);

}
;

/**
 * this is a hook that can be redefined by inheriting grammars to
 * introduce other sorts of binding behaviour for quantifiers or
 * substitution terms.  In particular, the Taclet parser will hook in
 * here to allow quantification of schema variables.
 */

bindingterm returns [Term a = null]
	:
    a = local_bindingterm             
    ;


/** here is the standard behaviour, that declares a new variable
  * with local scope in a quantifier */

local_bindingterm returns [Term a = null]
{
	Operator op = null;
	LogicVariable v=null;
	TermSymbol pv = null;
	Term a0=null;
	Term a1,a2;
	Sort s;
}
	:
		( 
			"eps" { op = Op.EPSILON; }
		|	"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		
		v0:IDENT COLON s=sort_id DOT {v=bindVar(v0,s);} a1=term60 
		{ 
			if ( op == Op.EPSILON ) {
				a = tf.createEpsilonTerm(v,a1);
			} else {
				a = tf.createQuantifierTerm((Quantifier)op, v,a1);
			}
			unbindVars();
		}
	|
		LBRACE
		( 
		  (term60 ASSIGN)=>
		  (a0=term60 ASSIGN a1=term RBRACE a2=term60
		      { 
			a = tf.createUpdateTerm(a0,a1,a2);			
		      }
		  ) | 		  
		  (v1:IDENT
			 ((ASSIGN a1=term RBRACE
			      {	pv = progVar(v1,a1.sort());} a2=term60
			      { 
				if (pv instanceof ProgramVariable) {
					a = tf.createUpdateTerm((ProgramVariable)pv,a1,a2);
				} else {
					a = tf.createUpdateTerm((SchemaVariable)pv,a1,a2);
				} 
			      })|
			 (COLON s=sort_id a1=term RBRACE 
			      {v = bindVar(v1,s);} a2=term60
			      { 
				a = tf.createSubstitutionTerm(Op.SUBST,v,a1,a2);
				unbindVars();
		      	      }
			  ))
      		       ) 
		 )
	;


/**
 * this is a hook that can be redefined by inheriting grammars to
 * redefine the sorts of things that can appear inside a diamond or
 * box.  In particular, in Taclets, schema variables and pre/suffixes
 * might occur.
 */

dl_term returns [Term a = null]
	:
		a = normal_dl_term
	;


/** here is the standard behaviour, only accepts a program. */

normal_dl_term returns [Term a = null]
{	JavaBlock jb=null;
	Term a1;
}
    :
		d1:DIAMOND 
		{
			jb=getJavaBlock(d1);
			HashSet progVars=progVars(jb);
			bindProgVars(progVars);
		}
		        a1=term60	
		{
			a = tf.createDiamondTerm(jb, a1);
			unbindProgVars();
		}
	|
		d2:BOX 
		{
			jb=getJavaBlock(d2);
			HashSet progVars=progVars(jb);
			bindProgVars(progVars);
		}
		        a1=term60	
		{
			a = tf.createBoxTerm(jb, a1);
			unbindProgVars();
		}
    ;


attrterm returns [Term a = null]
{	Vector vec=new Vector();
	ParsableVariable v=null;
	Term t=null;
	ParsableVariable v0=null;
    Sort so=null;

} : (t=funcpredvarterm { so=t.sort();
        } (DOT^ v=attrid[so] {vec.add(v); so=v.sort(); })+) {
	for (int i=0; i<vec.size(); i++) {
		t=tf.createAttributeTerm((ProgramVariable)vec.get(i),t.sort(),t);
	}
	a=t;
    };

attrid [Sort s] returns [ParsableVariable v = null]
{

} : attr:IDENT 
{ v = getJavaInfo().getAttribute(attr.getText(), (ObjectSort)s);};

funcpredvarterm returns [Term a = null]
{
	Operator vf;
	int n = 0;
	Vector al = null;
	Term t;
}
	:
		vf = varfunc_id  
		(  
			{ vf instanceof ParsableVariable }? 
			{ 	a = termForParsedVariable((ParsableVariable)vf); }
		|
			( { vf.arity() > 0 }?
				LPAREN 
				t=term 
				{
					al = new Vector();
					n=1;
					al.add(t);
				} 
				( COMMA 
					t=term 
					{
						n++;
						al.add(t);
					}
				)* RPAREN 
			)?
			{
			    if (vf instanceof Function) {
				if ( al == null ) {	
					a = tf.createFunctionTerm((Function)vf);
				} else {	
					a = tf.createFunctionTerm((Function)vf,
						   (Term[])al.toArray(AN_ARRAY_OF_TERMS));
				}
			    } else if (vf instanceof MetaOperator) {
				
				if ( al == null ) {	
					a = tf.createMetaTerm((MetaOperator)vf, new Term[0]);
				} else {	
					a = tf.createMetaTerm((MetaOperator)vf,
						   (Term[])al.toArray(AN_ARRAY_OF_TERMS));
				}
			    }
			}
		)
	;



varfunc_id returns [Operator v = null] 
{ 
	String varfunc_name = null;
}
	:
	(	v = expr_op	
		| varfunc_name = id
		{   
			v = (TermSymbol) var_ns.lookup(new Name(varfunc_name));
			if (v == null) {
				v= (Function) func_ns.lookup(new Name(varfunc_name));
				if (v == null) {
				  throw new NotDeclException
				      ("variable or function", varfunc_name, 
				       getFilename(), getLine(), getColumn());
				}
                
			}
		}
		| META_INT_UNARY_MINUS 
		{ v = Op.META_INT_UNARY_MINUS ; }
		| META_LONG_UNARY_MINUS 
		{ v = Op.META_LONG_UNARY_MINUS ; }
		| META_ADD 
		{ v = Op.META_ADD; }
		| META_SUB
		{ v = Op.META_SUB; }
		| META_MUL
		{ v = Op.META_MUL; }
		| META_DIV 
		{ v = Op.META_DIV; }
		| META_MOD
		{ v = Op.META_MOD; }
		| META_LESS
		{ v = Op.META_LESS; }
		| META_GREATER
		{ v = Op.META_GREATER; }
		| META_LEQ
		{ v = Op.META_LEQ; }
		| META_GEQ
		{ v = Op.META_GEQ; }
		| META_EQ
		{ v = Op.META_EQ; }
		| META_INT_ADD 
		{ v = Op.META_INT_ADD; }
		| META_INT_SUB 
		{ v = Op.META_INT_SUB; }
		| META_INT_MUL 
		{ v = Op.META_INT_MUL; }
		| META_INT_DIV 
		{ v = Op.META_INT_DIV; }
		| META_INT_MOD 
		{ v = Op.META_INT_MOD; }
		| META_INT_AND 
		{ v = Op.META_INT_AND; }
		| META_INT_OR
		{ v = Op.META_INT_OR; }
		| META_INT_XOR
		{ v = Op.META_INT_XOR; }
		| META_INT_SHIFTRIGHT
		{ v = Op.META_INT_SHIFTRIGHT; }
		| META_INT_SHIFTLEFT
		{ v = Op.META_INT_SHIFTLEFT; }
		| META_INT_UNSIGNEDSHIFTRIGHT
		{ v = Op.META_INT_UNSIGNEDSHIFTRIGHT; }
		| META_LONG_ADD 
		{ v = Op.META_LONG_ADD; }
		| META_LONG_SUB 
		{ v = Op.META_LONG_SUB; }
		| META_LONG_MUL 
		{ v = Op.META_LONG_MUL; }
		| META_LONG_DIV 
		{ v = Op.META_LONG_DIV; }
		| META_LONG_MOD 
		{ v = Op.META_LONG_MOD; }
		| META_LONG_AND 
		{ v = Op.META_LONG_AND; }
		| META_LONG_OR
		{ v = Op.META_LONG_OR; }
		| META_LONG_XOR
		{ v = Op.META_LONG_XOR; }
		| META_LONG_SHIFTRIGHT
		{ v = Op.META_LONG_SHIFTRIGHT; }
		| META_LONG_SHIFTLEFT
		{ v = Op.META_LONG_SHIFTLEFT; }
		| META_LONG_UNSIGNEDSHIFTRIGHT
		{ v = Op.META_LONG_UNSIGNEDSHIFTRIGHT; }
		| META_IN_LONG 
		{ v = Op.META_IN_LONG; }
		| META_IN_INT
		{ v = Op.META_IN_INT; }
		| META_IN_SHORT
		{ v = Op.META_IN_SHORT; }
		| META_IN_BYTE 
		{ v = Op.META_IN_BYTE; }
		| META_MODULO_BYTE
		{ v = Op.META_MODULO_BYTE; }
		| META_MODULO_SHORT
		{ v = Op.META_MODULO_SHORT; }
		| META_MODULO_INTEGER
		{ v = Op.META_MODULO_INTEGER; }
		| META_MODULO_LONG 
		{ v = Op.META_MODULO_LONG; }) 
	;


expr_op returns [Operator v = null] 
{
	ParsableVariable left_var0 = null;
	ParsableVariable right_var0 = null;
	ParsableVariable unary_var0 = null;
}	
	:

	IN_TYPE LPAREN ((left_var0=var_id ((id:ARITH_OP | id1:SLASH | id2:IDENT) right_var0 = var_id)?) |
			   ( "-" unary_var0 = var_id) )  RPAREN
	{  
           if (id==null) id=id1;	
	   if (unary_var0!=null) {
	      // can only be unary minus
	      SchemaVariable unary_var = (SchemaVariable) unary_var0;
	      Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
	      return new InTypeOperator
		(type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
	   }
	   else {
	      SchemaVariable left_var = (SchemaVariable) left_var0;
	      Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
	      if (right_var0 != null) {
	         SchemaVariable right_var = (SchemaVariable) right_var0;
    	         String op_str = (id == null ? id2.getText() : id.getText());
           	   if ("+".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
	         } else if ("-".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
	         } else if ("*".equals(op_str)) {
		     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
	         } else if ("/".equals(op_str)) {
	    	     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
		} else if ("%".equals(op_str)) {
	    	     return new InTypeOperator
		       (type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
  	         } 
	      }
	      else {
	         return new InTypeOperator(type, left_var);
	      }
 	   }
	}
;

var_id returns [ParsableVariable v = null]
	:
		id:IDENT 
		{   
			v = (ParsableVariable) var_ns.lookup(new Name(id.getText()));
			if (v == null) {
				throw new NotDeclException("variable", id, 
							   getFilename());
			}
		}
	;


