// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* manages the possible application of rules (RuleApps) */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.*;

public class RuleAppIndex implements java.io.Serializable {

    TacletAppIndex tacletAppIndex;

    public RuleAppIndex(TacletAppIndex tacletAppIndex) {
	this.tacletAppIndex=tacletAppIndex;
    }

    /**
     * returns the TacletApp index for this ruleAppIndex. 
     */
    public TacletAppIndex tacletAppIndex() {
	return tacletAppIndex;
    }

    /**
     * returns the Taclet index for this ruleAppIndex. 
     */
    public TacletIndex tacletIndex() {
	return tacletAppIndex.tacletIndex();
    }

     
    /**returns the set of rule applications for
     * the given heuristics 
     * at the given position of the given sequent.
     * @param heuristics the Heuristics that are enabled
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the PosInOccurrence to focus
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     */
    public ListOfTacletApp getTacletAppAt(ListOfHeuristic heuristics,
				      boolean interactive,
				      PosInOccurrence pos,
				      JavaInfo javaInfo) { 
	return tacletAppIndex.getTacletAppAt(heuristics, interactive, pos,
					 javaInfo); 
    }
    


    /**
     * returns the rule applications at the given PosInOccurrence and at all
     * Positions below this. The method calls getTacletAppAt for all the
     * Positions below.
     * @param heuristics the heuristics that are given to method getTacletApp
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the position where to start from
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the possible rule applications 
     */
    public ListOfTacletApp getTacletAppAtAndBelow(ListOfHeuristic heuristics,
					      boolean interactive,
					      PosInOccurrence pos,
					      JavaInfo javaInfo) {
	return tacletAppIndex.getTacletAppAtAndBelow(heuristics,
						 interactive, 
						 pos, javaInfo);
    }


    /** 
     * collects all FindTacletInstantiations for the given
     * heuristics and position
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getFindTaclet(boolean interactive,
					  PosInOccurrence pos,
					  ListOfHeuristic heuristics,
					  JavaInfo javaInfo) { 
	return tacletAppIndex.getFindTaclet(interactive, pos, heuristics, javaInfo);
    }

    /** 
     * collects all NoFindTacletInstantiations for the given
     * heuristics 
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp
	getNoFindTaclet(boolean interactive, 
		      ListOfHeuristic heuristics,
		      JavaInfo javaInfo) { 
	return tacletAppIndex.getNoFindTaclet(interactive, heuristics, 
					  javaInfo);	
    }


    /** 
     * collects all RewriteTacletInstantiations for the given
     * heuristics in a subterm of the constraintformula described by a
     * PosInOccurrence
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getRewriteTaclet
	(boolean interactive, PosInOccurrence pos,
	 ListOfHeuristic heuristics, JavaInfo javaInfo) { 
	return tacletAppIndex.getRewriteTaclet(interactive, pos,
					   heuristics, javaInfo);
    }


    /**
     * adds a new Taclet with instantiation information to the Taclet Index 
     * of this TacletAppIndex.
     * @param tacletApp the partially instantiated Taclet to add
     */
    public void addNoPosTacletApp(NoPosTacletApp tacletApp) {
	tacletAppIndex.addNoPosTacletApp(tacletApp);
    }


    /** called if a formula has been added
     * @param addedCF the added ConstrainedFormula
     * @param p PosInOccurrence position of the added formula 
     * @param before a boolean true iff the formula has been added
     * before the one decribed in p
     */  
    public void formulaAdded(ConstrainedFormula addedCF,
			     PosInOccurrence p) {
	
    }

    /** called if a formula has been replaced
     * @param changedCF the ConstrainedFormula that replaced the old one
     * @param p PosInOccurrence with position of the formula that has
     * been replaced
     */  
    public void formulaChanged(ConstrainedFormula changedCF, PosInOccurrence p){
    }

    /** called if a formula has been removed
     * @param p PosInOccurrence with former position of the formula
     * that has been removed.
     */  
    public void formulaRemoved(PosInOccurrence p) {
    }


    /**
     * returns a new RuleAppIndex with a copied Taclet Index
     */
    public RuleAppIndex copy() {
	return new RuleAppIndex(tacletAppIndex().copy());
    }


    public String toString() {
	return "RuleAppIndex without indexing, getting Taclets from"
	    +" TacletAppIndex "+tacletAppIndex;
    }

    /**
     * @link aggregation
     * @supplierCardinality 1
     * @clientCardinality 1 
     */
    /*#private TacletAppIndex lnkTacletAppIndex;*/

    /**
     * @clientCardinality 1
     * @clientQualifier possible rule applications
     * @supplierCardinality 1 
     */
    /*#Goal lnkGoal;*/

    /** @link aggregation 
     * @supplierCardinality *
     * @clientCardinality 1
     * @supplierRole managed rule applications*/
    /*#RuleApp lnkRuleApp;*/
}
