// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* Generated by Together */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.SkolemFunction;
import java.util.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;

/**
 * A skolem deliverer instance is responsible for the management of skolemsymbols
 * within a proof. It keeps track of the created skolemsymbols for each sort and
 * gives skolemsymbols not occuring in the current proof(environment).
 * For each proof there has to be a seperate skolem deliverer !
 */

public class SkolemDeliverer {
    /** 
     * For each (used) sort there will be created a counter with the number of the 
     * skolemsymbol (of that sort) last generated.
     */
    private Map counterForSorts; // from Sort to Integer ...
    
    public SkolemDeliverer(){
	counterForSorts = new HashMap();
    }

    /**
     * Delivers a skolemsybol which is of the given sort and new in the context of the
     * proof for which this skolem deliverer was used. More exactly the returned symbol
     * has never been created by this skolem deliverer before !
     * @param sort - sort of the skolem function
     * @param argSorts - sorts of the arguments of the skolem function
     * @returns a new skolemsymbol of the given sort
     */
    public SkolemFunction getSkolemSymbolForSort(Sort sort, Sort[] argSorts){
	// Check whether there is an entry for the given sort, create one if necessary ...
    TermFactory tf = TermFactory.DEFAULT;
    SkolemFunction sft;
    
 	if(counterForSorts.containsKey(sort)){
	    // enlarge counter ...
	    Integer cnt = (Integer) counterForSorts.get(sort);
	    Integer newCnt = new Integer(cnt.intValue() + 1);
	    counterForSorts.put(sort,newCnt);
	    cnt = null;
	    sft = new SkolemFunction(new Name(sort.toString() +
					      SkolemFunction.SKOLEM_LABEL +
					      newCnt), sort, argSorts);
	} else {
	    // create new entry ...
	    Integer newCnt = new Integer(0);
	    counterForSorts.put(sort,newCnt);
	    sft = new SkolemFunction(new Name(sort.toString() +
					      SkolemFunction.SKOLEM_LABEL +
					      newCnt), sort, argSorts);
	}
	
	return sft;
    }

     /**
     * Delivers a skolemsybol which is of the given sort and new in the context of the
     * proof for which this skolem deliverer was used. More exactly the returned symbol
     * has never been created by this skolem deliverer before !
     * @param varToSkolemize - variable to be skolemized; determines the name and sort
     *                         of the skolem function.
     * @param argSorts - sorts of the arguments of the skolem function
     * @returns a new skolemsymbol of the given sort
     */
    
    public SkolemFunction getSkolemSymbolForVariable(QuantifiableVariable varToSkolemize, Sort[] argSorts){
	// Check whether there is an entry for the given sort, create one if necessary ...
    TermFactory tf = TermFactory.DEFAULT;
    SkolemFunction sft;
    Sort sort = varToSkolemize.sort();
    
 	if(counterForSorts.containsKey(sort)){
	    // enlarge counter ...
	    Integer cnt = (Integer) counterForSorts.get(sort);
	    Integer newCnt = new Integer(cnt.intValue() + 1);
	    counterForSorts.put(sort,newCnt);
	    cnt = null;
	    sft = new SkolemFunction(new Name(varToSkolemize.name().toString() +
					      SkolemFunction.SKOLEM_LABEL +
					      newCnt), sort, argSorts);
	} else {
	    // create new entry ...
	    Integer newCnt = new Integer(0);
	    counterForSorts.put(sort,newCnt);
	    sft = new SkolemFunction(new Name(varToSkolemize.name().toString() +
					      SkolemFunction.SKOLEM_LABEL +
					      newCnt), sort, argSorts);
	}
	
	return sft;
    }

}
