// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.init;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.gui.PresentationFeatures;
import de.uka.ilkd.key.logic.*;

public class ProblemInitializer {


    private Namespace var_ns  = new Namespace();
    private Namespace func_ns = new Namespace();
    private Namespace sort_ns = new Namespace();
    private Namespace heur_ns = new Namespace();

    private Main main = null; // for status reporting -- optional

    private ProofOblInput problem = null;
    private EnvInput[] modelDepInputs=null;
    private EnvInput[] modelIndepInputs=null;
    private boolean standardInit;

    private static final RuleSource PROP_RULES = 
       RuleSource.initRuleFile("propRule.key"); 

    private static final RuleSource SORT_RULES = 
       RuleSource.initRuleFile("sortRule.key");

    private static final RuleSource JAVA_RULES = 
       RuleSource.initRuleFile("javaRules.key"); 

    public static InitConfig INIT_CONFIG=null;

    private InitConfig initConfig=null;

    private LDTInput ldtIn;
    
    public ProblemInitializer(ProofOblInput problemFile, ProofSettings settings,
                              Main main) {
       this.main = main;
       this.problem=problemFile;
       if (INIT_CONFIG==null) {
	   initConfig=new InitConfig();
	   standardInit=false;
       } else {
	   initConfig=INIT_CONFIG.copy();
	   standardInit=true;
       }
       modelIndepInputs
	   = new EnvInput[] {new LDTInput(settings.getLDTSettings(), main), 
			     new KeYFile("propositional taclets", PROP_RULES, main),
			     new KeYFile("Java taclets", JAVA_RULES, main)};
       ldtIn=(LDTInput)modelIndepInputs[0];
       modelDepInputs 
	   = new EnvInput[] {new SortKeYFile("sort taclets", SORT_RULES, main)};
    }

    public InitConfig initConfig() {
	return initConfig;
    }

    private void readEnvInputs(EnvInput[] in) {
	for (int i=0; i<in.length; i++) {
	    if (main != null) {
		int chars = in[i].getNumberOfChars();
		main.setStatusLine("Reading "+in[i].name(), chars);
	    }
	    in[i].addTo(initConfig);
	    in[i].read(ModStrategy.NO_VARS);
	}
    }
       
    public void forceInitializeStandardRules() {
	readEnvInputs(modelIndepInputs);
    }

    public void initializeStandardRules() {
	if (!standardInit) {
	    forceInitializeStandardRules();
	    standardInit=true;
	    INIT_CONFIG=initConfig.copy();
	}
    }

    public void initializeJavaModel() {
	if (main != null) {
	    main.setStatusLine("Reading Java model");
	}
	problem.setConfig(initConfig);
	problem.readJava();
    }

    public POInputResult initializeProblem() {
	if (main != null) 
           main.setStatusLine("Loading problem \""+problem.name()+"\"");
	problem.setConfig(initConfig);
	problem.read(ModStrategy.MOD_ALL);
	readEnvInputs(modelDepInputs);
	return problem.getProblemTerm();
    }

    public void setUpTest() {
	initializeJavaModel();
	POInputResult f = initializeProblem();
	InitConfig pi=initConfig();	  
    }

    public void setUpProof() {
	initializeJavaModel();
	POInputResult f = initializeProblem();
	InitConfig pi=initConfig();	  
	for (int i=0; i<f.size(); i++) {
	    main.addProblem(f.name(i), f.term(i), pi.createTacletIndex(),
			    pi.varNS(), pi.funcNS(), pi.sortNS(), pi.heurNS(),
			    pi.javaInfo());		
	}	
	main.setStandardStatusLine();
	PresentationFeatures.initialize(main.mediator().func_ns(), 
					main.mediator().getNotationInfo(),
					main.mediator().getSelectedProof());
    }

    public void tryReadProof(ProblemLoader prl) throws antlr.ANTLRException {
	if (problem instanceof KeYFile) {
	    ((KeYFile)problem).readProof(prl);
	}
    }

    public LDT[] ldtModels() {
	return ldtIn.ldtModels();
    }

    private void add(Namespace target, Namespace source) {
	IteratorOfNamed it=source.elements().iterator();
	while (it.hasNext()) {
	    target.add(it.next());
	}
	
    }

}
