// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** this class wrapps a MapFromSchemaVariableToInstantiationEntry and is used to
 * store instantiations of schemavariables. The class is immutable,
 * this means changing its content will result in creating a new object.
 */
package de.uka.ilkd.key.rule.inst;

import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.ArrayOfProgramElement;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;
import java.io.Serializable;
import java.io.ObjectStreamException;
import de.uka.ilkd.key.logic.*;



public class SVInstantiations implements java.io.Serializable {
    /** the empty instantiation */
    public final static SVInstantiations EMPTY_SVINSTANTIATIONS 
	= new SVInstantiations();

    /** the map with the instantiations to logic terms*/
    private MapFromSchemaVariableToInstantiationEntry map = 
	MapAsListFromSchemaVariableToInstantiationEntry.EMPTY_MAP;

    private final static SchemaVariable CONTEXTSV
	= SchemaVariable.createProgramSV(new Name("Context"), 
					 new ProgramSVSort(new Name
					     ("ContextTerm"))); 
                                                  //just a dummy SV for context

    private ListOfUpdatePair updateContext = SLListOfUpdatePair.EMPTY_LIST;

    /** creates a new SVInstantions object with an empty map */
    private SVInstantiations() {
    }
    
    /** creates a new SVInstantions object using the given map 
     * @param map the MapFromSchemaVariableToInstantiationEntry with the
     * instantiations
     */
    private SVInstantiations(MapFromSchemaVariableToInstantiationEntry map,
			     ListOfUpdatePair updateContext) {
	this.map = map;
	this.updateContext = updateContext;
    }

    /** */
    private SVInstantiations(UpdateOp op, Term update) {
	updateContext = updateContext.append(new UpdatePair(op, update));
    }


    private Object readResolve() 
	throws ObjectStreamException {
	if (map == MapAsListFromSchemaVariableToInstantiationEntry.EMPTY_MAP) {
	    return EMPTY_SVINSTANTIATIONS;
	}
	return this;
    }

    /** adds the given pair to the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return SVInstantiations the new SVInstantiations containing
     * the given pair
     */
    public SVInstantiations add(SchemaVariable sv, Term subst) {
	return new SVInstantiations(map.put
				    (sv, 
				     new TermInstantiation(sv, subst)),
				    getUpdateContext());
    }    

    /** adds the given pair to the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param pe the ProgramElement the SchemaVariable is instantiated with
     * @return SVInstantiations the new SVInstantiations containing
     * the given pair
     */
    public SVInstantiations add(SchemaVariable sv, ProgramElement pe) {
	return new SVInstantiations(map.put
				    (sv, 
				     new ProgramInstantiation(sv, pe)),
				    getUpdateContext());
    }    


    public SVInstantiations add(SchemaVariable sv, ArrayOfProgramElement pes) {
	return new SVInstantiations(map.put
				    (sv, 
				     new ProgramListInstantiation(sv, pes)),
				    getUpdateContext());
    }

    /** adds the given pair to the instantiations for the context.If the context 
     * has been instantiated already, the new pair is
     * taken without a warning.
     * @param prefix the PosInProgram describing the prefix
     * @param postfix the PosInProgram describing the postfix
     * @return SVInstantiations the new SVInstantiations containing
     * the given pair
     */
    public SVInstantiations add(PosInProgram prefix,
				PosInProgram postfix,
				ProgramElement pe) {
	return new SVInstantiations
	    (map.put(CONTEXTSV, 
		     new ContextInstantiationEntry(CONTEXTSV, prefix,
						   postfix, pe)),
	     getUpdateContext());
    }    




    /** adds the given pair to the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param entry the InstantiationEntry
     * @return SVInstantiations the new SVInstantiations containing
     * the given pair
     */
    public SVInstantiations add(SchemaVariable sv, 
				InstantiationEntry entry) {
	return new SVInstantiations(map.put(sv, entry), 
				    getUpdateContext());
    }    

    /** replaces the given pair in the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param entry the InstantiationEntry the SchemaVariable is 
     * instantiated with
     */
    public SVInstantiations replace(SchemaVariable sv, 
				    InstantiationEntry entry) {
	return new SVInstantiations(map.remove(sv).put(sv, entry),
				    getUpdateContext());
    }    

    /** replaces the given pair in the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is 
     * instantiated with
     */
    public SVInstantiations replace(SchemaVariable sv, 
				    Term term) {
	return replace(sv, new TermInstantiation(sv, term));
    }    

    /** replaces the given pair in the instantiations. If the given
     * SchemaVariable has been instantiated already, the new pair is
     * taken without a warning.
     * @param sv the SchemaVariable to be instantiated
     * @param pe the ProgramElement the SchemaVariable is 
     * instantiated with
     */
    public SVInstantiations replace(SchemaVariable sv, 
				    ProgramElement pe) {
	return replace(sv, new ProgramInstantiation(sv, pe));
    }    
   
    /** replaces the given pair in the instantiations. If the context
     * has been instantiated already, the new pair is
     * taken without a warning.
     * @param prefix the PosInProgram describing the position of the
     * first statement after the prefix
     * @param postfix the PosInProgram describing the position of the
     * statement just before the postfix
     */
    public SVInstantiations replace(PosInProgram prefix,
				    PosInProgram postfix, ProgramElement pe) {
	return replace(CONTEXTSV, new ContextInstantiationEntry(CONTEXTSV, 
							 prefix, 
							 postfix,pe));
    }    
   
 
    /** returns true iff the sv has been instantiated already 
     * @return  true iff the sv has been instantiated already 
    */
    public boolean isInstantiated(SchemaVariable sv) {
	return map.containsKey(sv);
    }

    /** returns the instantiation of the given SchemaVariable
     * @return the InstantiationEntry the SchemaVariable will be instantiated with, 
     * null if no instantiation is stored
    */
    public InstantiationEntry getInstantiationEntry(SchemaVariable sv) {
	return map.get(sv);
    }

    /** returns the instantiation of the given SchemaVariable
     * @return the Object the SchemaVariable will be instantiated with, 
     * null if no instantiation is stored
    */
    public Object getInstantiation(SchemaVariable sv) {
	InstantiationEntry entry = getInstantiationEntry(sv);
	return entry == null ? null : entry.getInstantiation();
    }

    /** adds an update to the update context */
    public SVInstantiations addUpdate(UpdateOp op, Term update) {
	return new SVInstantiations(map, updateContext.append
				    (new UpdatePair(op, update)));
    }

    public SVInstantiations addUpdateList(ListOfUpdatePair updates) {
	return new SVInstantiations(map, updates);
    }

    /**
     * returns the instantiation entry for the context "schema variable" or
     * null if non such exists
     */
    public ContextInstantiationEntry getContextInstantiation() {
	InstantiationEntry entry = getInstantiationEntry(CONTEXTSV);
	return (ContextInstantiationEntry) entry;
    }

    /** returns iterator of the SchemaVariables that have an
     * instantiation
     * @return the IteratorOfSchemaVariable
     */
    public IteratorOfSchemaVariable svIterator() {
	return map.keyIterator();
    }

    /** returns iterator of the mapped pair (SchemaVariables, InstantiationEntry)
     * @return the IteratorOfEntryOfSchemaVariableAndInstantiationEntry
     */
    public IteratorOfEntryOfSchemaVariableAndInstantiationEntry pairIterator() {
	return map.entryIterator();
    }

    /** returns the number of SchemaVariables of which an
     * instantiation is known
     * @return int that is the number of SchemaVariables of which an
     * instantiation is known
     */
    public int size() {
	return map.size();
    }

    /** returns the update context 
     * @return the update context 
     */
    public ListOfUpdatePair getUpdateContext() {
	return updateContext;
    }

    /** returns true if the given object and this one have the same
     * mappings
     * @return true if the given object and this one have the same
     * mappings
     */ 
    public boolean equals(Object obj) {
	SVInstantiations cmp = null;
	if (!(obj instanceof SVInstantiations)) {
	    return false;
	} else {
	    cmp = (SVInstantiations) obj;
	}
	if (!getUpdateContext().
	    equals(cmp.getUpdateContext())) {
	    return false;
	}
	if (size() != cmp.size()) {
	    return false;
	} else {
	    IteratorOfSchemaVariable it = svIterator();
	    while (it.hasNext()) {
		SchemaVariable sv = it.next();
		if (!getInstantiation(sv).equals
		    (cmp.getInstantiation(sv))) {
		    return false;
		}
	    }
	    return true;
	}
    }

    public SVInstantiations union(SVInstantiations other) {
	MapFromSchemaVariableToInstantiationEntry result = map;
	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it = 
	    other.map.entryIterator();
	while (it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry entry = it.next();	    
	    result = result.put(entry.key(), entry.value());	    
	}
	ListOfUpdatePair updates = SLListOfUpdatePair.EMPTY_LIST;
	if (other.getUpdateContext() == SLListOfUpdatePair.EMPTY_LIST) {
	    updates = getUpdateContext();
	} else if (getUpdateContext() == SLListOfUpdatePair.EMPTY_LIST) {
	    updates = other.getUpdateContext();
	} else if (!getUpdateContext().equals(other.getUpdateContext())) {
	    Debug.assert(false,
			 "The update context of one of"+
			 " the instantiations has to be empty or equal.");
	} else {
	    updates = other.getUpdateContext();
	}
	return new SVInstantiations(result, updates);	
    }

    /** toString */
    public String toString() {
	StringBuffer result = new StringBuffer("SV Instantiations: ");
	return (result.append(map.toString())).toString();
    }


    static class InstantiationException extends RuntimeException{
    
	public InstantiationException (String msg) {
	    super(msg);
	}
	
    }

}
